import glob
import logging
import os
import pprint
import re
import traceback
import unittest
from pathlib import Path

import lxml
from lxml import etree
from lxml.etree import XMLSyntaxError, _Element

# local
from amilib.ami_dict import AmiDictionary, AmiEntry, AMIDictError, \
    AmiDictValidator, NAME, TITLE, TERM, LANG_UR, VERSION, WIKIDATA_ID
from amilib.amix import AmiLib
from amilib.constants import LOCAL_CEV_OPEN_DICT_DIR
from amilib.dict_args import AmiDictArgs
from amilib.dict_args import MARKUP_FILE
from amilib.file_lib import FileLib
from amilib.util import Util
from amilib.wikimedia import WikidataSparql
from amilib.xml_lib import XmlLib
from amilib.ami_html import HtmlLib
from test.resources import Resources
from test.test_all import AmiAnyTest

# from parametrized import parametrized

# MUST use RAW content , not HTML
CEV_OPEN_RAW_DICT_URL = "https://raw.githubusercontent.com/petermr/CEVOpen/master/dictionary/"
PLANT_PART_RAW_DICT_URL = CEV_OPEN_RAW_DICT_URL + "eoPlantPart/eoplant_part.xml"
COMPOUND_RAW_DICT_URL = CEV_OPEN_RAW_DICT_URL + "eoCompound/plant_compound.xml"
ANALYSIS_METHOD_RAW_DICT_URL = CEV_OPEN_RAW_DICT_URL + "eoAnalysisMethod/eoAnalysisMethod.xml"
TEST_DIR = Path(Path(__file__).parent.parent, "test")
TEST_RESOURCE_DIR = Path(TEST_DIR, "resources")
DICTFILE1 = "dictfile1"
ROOT = "root"
ONE_ENTRY_PATH = "one_entry_file"
ONE_ENTRY_DICT = "one_entry_dict"
MINI_PLANT_PART = "mini_plant_part"
MINI_MENTHA = "mini_mentha"
ETHNOBOT_DICT = "VC_EthnobotanicalUse"
DUPLICATE_ENTRIES = "test_duplicate_entries"

RESOURCES_DIR = "resources"
# AMIDICTS = Path(Path(__file__).parent.parent, "py4ami/resources/amidicts")  # relative to dictribution base
AMIDICTS = Path(Path(__file__).parent.parent, f"{RESOURCES_DIR}/amidicts")  # relative to dictribution base

STARTING_VERSION = "0.0.1"

HOME = os.path.expanduser("~")
IPCC_DICT_ROOT = Path(HOME, "projects/semanticClimate/ipcc/ar6/wg3")  # needs resetting if tests are to be run

logger = Util.get_logger(__name__, level=logging.DEBUG)

# ===== helpers =====
def _create_amidict_with_foo_bar_entries():
    amidict = AmiDictionary.create_minimal_dictionary()
    entry_foo = amidict.create_and_add_entry_with_term("foo")
    entry_bar = amidict.create_and_add_entry_with_term("bar")
    return amidict


class AmiDictionaryTest(AmiAnyTest):
    """These are tests for developing CODE for dictionary creation and validation

    Code for VALIDATION of dictionaries should probably be bundled with the dictionaries themselves

    """

    logging.info(f"loading {__file__}")

    HTML_WITH_IDS = "html_with_ids"

    DICTFILE1 = "dictfile1"
    ROOT = "root"
    ONE_ENTRY_PATH = "one_entry_file"
    ONE_ENTRY_DICT = "one_entry_dict"
    MINI_PLANT_PART = "mini_plant_part"
    MINI_MENTHA = "mini_mentha"
    ETHNOBOT_DICT = "VC_EthnobotanicalUse"
    DUPLICATE_ENTRIES = "test_duplicate_entries"

    # AMIDICTS = Path(Path(__file__).parent.parent, "py4ami/resources/amidicts")  # relative to dictribution base
    # AMIDICTS = Path(Path(__file__).parent.parent, f"{AMI_TOP}/resources/amidicts")  # relative to dictribution base

    STARTING_VERSION = "0.0.1"

    ADMIN = True and AmiAnyTest.ADMIN
    CMD = True and AmiAnyTest.CMD
    LONG = True and AmiAnyTest.LONG
    NET = True and AmiAnyTest.NET
    NYI = True and AmiAnyTest.NYI
    USER = True and AmiAnyTest.USER
    VERYLONG = True and AmiAnyTest.VERYLONG

    CEV_EXISTS = Path(LOCAL_CEV_OPEN_DICT_DIR).exists()

    def create_file_dict(self):
        """Variables created afresh for every test"""
        dictfile1 = Path(AMIDICTS, "dict1.xml")
        # dictfile1 = Path(Path(__file__).parent.parent, "py4ami/resources/amidicts/dict1.xml")
        one_entry_path = Path(Path(__file__).parent.parent, f"{AMIDICTS}/dict_one_entry.xml")
        root = etree.parse(str(dictfile1)).getroot()
        assert dictfile1.exists(), "{dictfile1} exists"
        one_entry_path = Path(AMIDICTS, "dict_one_entry.xml")
        one_entry_dict_new = AmiDictionary.create_from_xml_file(one_entry_path)
        assert one_entry_dict_new is not None
        mini_plant_part_path = Path(AMIDICTS, "mini_plant_part.xml")

        # BUG: this should be available through pytest
        setup_dict = {
            DICTFILE1: dictfile1,  # type path
            ROOT: root,
            ONE_ENTRY_PATH: one_entry_path,
            ONE_ENTRY_DICT: one_entry_dict_new,
            MINI_PLANT_PART: mini_plant_part_path,
            MINI_MENTHA: Path(AMIDICTS, "mentha_tps.xml"),
            ETHNOBOT_DICT: Path(AMIDICTS, ETHNOBOT_DICT + ".xml"),
            DUPLICATE_ENTRIES: Path(AMIDICTS, DUPLICATE_ENTRIES + ".xml"),
        }
        logger.info(f"setup_dict {setup_dict}")
        return setup_dict

    @unittest.skipUnless("environment", ADMIN)
    def test_dictionary_file1_exists(self):
        """Test that a simple dictionary "dictfile1" file exists"""
        setup_dict = self.create_file_dict()
        assert setup_dict[DICTFILE1].exists(), f"file should exist {setup_dict['dict1']}"

    def test_dictionary_element(self):
        """
        reads xml string into dictiomary and tests that has "dictionary" root element
        and no child entries
        """
        dict_str = """
        <dictionary title='foo'>
        </dictionary>
        """
        ami_dict = AmiDictionary.create_dictionary_from_xml_string(dict_str)
        assert ami_dict is not None
        assert ami_dict.root.tag == "dictionary"
        assert ami_dict.has_valid_root_tag()
        assert ami_dict.get_entry_count() == 0


    def test_title_from_url_stem(self):
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.url = "https://some.where/foo/bar.xml"
        assert amidict.root.attrib[TITLE] == "minimal"  # needs fixing

    def test_title_from_file_stem(self):
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.file = "/user/me/foo.xml"
        assert amidict.root.attrib[TITLE] == "minimal"

    def test_dict_has_root_dictionary(self):
        """
        Tests that the dictionary has <dictionary> root element
        """
        setup_dict = self.create_file_dict()
        root = setup_dict[ROOT]
        assert root.tag == AmiDictionary.TAG

    def test_dict_contains_xml_element(self):
        root = etree.parse(str(self.create_file_dict()[DICTFILE1]))
        assert root is not None

    def test_can_read_dictionary_from_url_as_xml(self):
        """
        Checks that a dictionary can be read from a URL into XML
        Reads PLANT_PART_RAW_DICT_URL and validates that has about 728 entries
        """

        url = PLANT_PART_RAW_DICT_URL
        tree = XmlLib.parse_url_to_tree(url)
        descendants = tree.getroot().xpath('.//*')
        assert 730 >= len(descendants) >= 720

    def test_dictionary_has_xml_declaration_with_encoding(self):
        """Checks dictionary has encoding of 'UTF-8' and XML Version 1.0
        USEFUL 2022-07"""
        dicts = [ETHNOBOT_DICT, DICTFILE1, ]
        for dikt in dicts:
            logger.debug(f"...{dikt}")
            root = etree.parse(str(self.create_file_dict()[dikt]))
            dictionary = AmiDictionary.create_from_xml_object(root)
            validator = AmiDictValidator(dictionary)
            error_list = validator.get_xml_declaration_error_list()
            assert not error_list

    # @pytest.mark.url
    def test_validate_url_dict(self):
        """
        tests that historic dictionaries read into validator
        TODO skip this
        """
        urllist = [
            PLANT_PART_RAW_DICT_URL,
            ANALYSIS_METHOD_RAW_DICT_URL,
            COMPOUND_RAW_DICT_URL
        ]
        for url in urllist:
            logger.debug(f"url: {url}")
            # tree = XmlLib.parse_url_to_tree(url)
            # dictionary = AmiDictionary.create_from_xml_object(tree)
            dictionary = AmiDictionary.create_dictionary_from_url(url)
            validator = AmiDictValidator(dictionary)
            # validator.validate_title()
            error_list = validator.get_error_list()
            logger.warning(f"error_list {error_list}")

    # AmiDictionary

    def test_can_create_ami_dict_from_file(self):
        """read an existing XML AmiDictionary"""
        setup_dict = self.create_file_dict()
        one_entry_path = setup_dict[ONE_ENTRY_PATH]
        amidict = AmiDictionary.create_from_xml_file(one_entry_path)
        assert amidict is not None

    # @pytest.mark.simple
    def test_dictionary_is_an_ami_dictionary(self):
        """
        test dictionary with one entry
        """
        setup_dict = self.create_file_dict()
        amidict = setup_dict[ONE_ENTRY_DICT]
        assert type(amidict) is AmiDictionary

    def test_dictionary_get_entries(self):
        """
        unit test
        """
        setup_dict = self.create_file_dict()
        amidict = setup_dict[ONE_ENTRY_DICT]
        entries = amidict.get_lxml_entries()
        assert entries is not None

    # @pytest.mark.simple
    def test_dictionary_contains_one_entry(self):
        """
        unit test
        """
        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert amidict.get_entry_count() == 1, f"dict should have 1 entry, found  {amidict.get_entry_count()}"

    def test_get_first_entry(self):
        """
        unit test
        """
        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert amidict.get_first_ami_entry() is not None

    def test_get_attribute_names(self):
        """
        unit test
        """

        first_entry = self.create_file_dict()[ONE_ENTRY_DICT].get_first_ami_entry()
        assert type(first_entry) is AmiEntry
        attrib_names = {name for name in first_entry.element.attrib}
        assert attrib_names is not None

    def test_get_term_of_first_entry(self):
        """
        tests that we can retrieve the `term` value from an element

        """
        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert amidict.get_first_entry().attrib[TERM] == "Douglas Adams"

    def test_get_name_of_first_entry(self):
        """
        unit test
        """

        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert amidict.get_first_entry().attrib[NAME] == "Douglas Adams"

    def test_get_wikidata_of_first_entry(self):
        """
        unit test
        """

        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert amidict.get_first_entry().attrib[WIKIDATA_ID] == "Q42"

    def test_get_synonym_count(self):
        """
        unit test
        """
        amidict = AmiDictionaryTest().create_file_dict()[ONE_ENTRY_DICT]
        assert type(amidict) is AmiDictionary
        assert len(amidict.get_first_ami_entry().get_synonyms()) == 2

    def test_get_synonym_by_language(self):
        """
        unit test
        """

        amidict = self.create_file_dict()[ONE_ENTRY_DICT]
        assert type(amidict) is AmiDictionary
        elem = amidict.get_first_ami_entry().get_synonym_by_language(LANG_UR).element
        assert "ڈگلس ایڈمس" == ''.join(elem.itertext())

    def test_add_entry_with_term_to_zero_entry_dict(self):
        """
        Creates minimal dictionary with no entries and adds one entry
        tests its components
        """
        amidict = AmiDictionary.create_minimal_dictionary()
        print()
        logger.info(f"amidict {amidict.get_entry_count()}")
        entry = amidict.create_and_add_entry_with_term("foo")
        assert etree.tostring(entry) == b'<entry term="foo"/>'
        assert etree.tostring(
            amidict.root) == b'<dictionary title="minimal" version="0.0.1"><entry term="foo"/></dictionary>'
        assert amidict.get_entry_count() == 1

    def test_add_two_entry_with_term_to_zero_entry_dict(self):
        amidict = AmiDictionary.create_minimal_dictionary()
        entry_foo = amidict.add_entry_element("foo")
        entry_bar = amidict.add_entry_element("bar")
        assert etree.tostring(entry_bar) == b'<entry name="bar" term="bar"/>'
        assert etree.tostring(
            amidict.root) == b'<dictionary title="minimal" version="0.0.1"><entry name="foo" term="foo"/><entry name="bar" term="bar"/></dictionary>'
        assert amidict.get_entry_count() == 2

    def test_add_list_of_entries_from_list_of_string(self):
        """
        from a list of strings creates a list of entries and adds to existing dictionary
        (the entries only have term/name fields)
        """
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        term_count = len(terms)
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.add_entries_from_words(terms)
        assert amidict.get_entry_count() == term_count

    def test_find_entry_after_add_list_of_entries_from_list_of_string(self):
        """
        creates entries from strings and tests that they can be accessed by term
        """
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.add_entries_from_words(terms)
        entry_bar = amidict.get_lxml_entry("bar")
        assert entry_bar is not None

    def test_fail_on_missing_entry_after_add_list_of_entries_from_list_of_string(self):
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.add_entries_from_words(terms)
        entry_zilch = amidict.get_lxml_entry("zilch")
        assert entry_zilch is None, f"missing entry returns None"

    def test_add_second_list_of_entries_from_list_of_string(self):
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.add_entries_from_words(terms)
        terms1 = ["wibble", "wobble"]
        amidict.add_entries_from_words(terms1)
        assert amidict.get_entry_count() == len(terms) + len(terms1)

    def test_add_list_of_entries_from_list_of_string_with_duplicates_and_replace(self):
        """
        creates dictionary from list of words and force replacement of duplicates
        """
        terms = ["foo", "bar", "plugh", "xyzzy", "bar"]
        amidict = AmiDictionary.create_minimal_dictionary()
        amidict.add_entries_from_words(terms, duplicates="replace")
        assert amidict.get_entry_count() == 4, f"'bar' should be present"

    def test_add_list_of_entries_from_list_of_string_with_duplicates_and_no_replace(self):
        """
        add list of terms which contains duplicate and raise error
        """
        terms = ["foo", "bar", "plugh", "xyzzy", "bar"]
        amidict = AmiDictionary.create_minimal_dictionary()
        try:
            amidict.add_entries_from_words(terms, duplicates="error")
            assert False, f"AMIDict duplicate error (bar) should have been thrown"
        except AMIDictError:
            assert True, "error should have been throwm"
        assert amidict.get_entry_count() == 4, f"'bar' should be present"

    def test_add_then_remove_entry_and_replace(self):
        """create new entry , then delete, then re-add"""
        amidict, _ = AmiDictionary.create_dictionary_from_words(["foo", "bar", "plugh", "xyzzy"])
        assert amidict.get_entry_count() == 4
        amidict.delete_entry_by_term("bar")
        assert amidict.get_entry_count() == 3, f"entry 'bar' should have been removed"
        amidict.create_and_add_entry_with_term("bar")
        assert amidict.get_entry_count() == 4, f"entry 'bar' should have been re-added"

    # find entries
    def test_find_entry_by_term(self):
        """searches for entry by value of term"""
        amidict = _create_amidict_with_foo_bar_entries()
        entry = amidict.get_lxml_entry("foo")
        assert entry is not None
        assert entry.attrib[TERM] == "foo", f"should retrieve entry with term 'foo'"

    def test_find_entry_by_term_bar(self):
        amidict = _create_amidict_with_foo_bar_entries()
        entry = amidict.get_lxml_entry("bar")
        assert entry is not None

    def test_find_entry_by_term_zilch(self):
        amidict = _create_amidict_with_foo_bar_entries()
        entry = amidict.get_lxml_entry("zilch")
        assert entry is None

    def test_delete_entry_by_term_foo(self):
        amidict = _create_amidict_with_foo_bar_entries()
        logger.debug(f"amidict0 {lxml.etree.tostring(amidict.root)}")
        amidict.delete_entry_by_term("foo")
        logger.debug(f"amidict1 {lxml.etree.tostring(amidict.root)}")
        assert amidict.get_entry_count() == 1

    def test_delete_entry_by_term_foo_and_re_add(self):
        amidict = _create_amidict_with_foo_bar_entries()
        amidict.delete_entry_by_term("foo")
        amidict.create_and_add_entry_with_term("foo")
        assert amidict.get_entry_count() == 2

    def test_create_and_add_entry_with_term(self):
        term = "foo"
        amidict = AmiDictionary.create_minimal_dictionary()
        assert amidict.get_entry_count() == 0
        amidict.create_and_add_entry_with_term(term)
        assert amidict.get_entry_count() == 1
        entry = amidict.get_ami_entry(term)
        assert type(entry) is AmiEntry
        assert term == entry.get_term()

    def test_create_and_overwrite_entry_with_duplicate_term(self):
        term = "foo"
        amidict = AmiDictionary.create_minimal_dictionary()
        assert amidict.get_entry_count() == 0
        entry = amidict.create_and_add_entry_with_term(term)
        logger.debug(f"entry: {type(entry)}")
        assert isinstance(entry, _Element)
        AmiEntry.add_name(entry, "foofoo")
        amidict.create_and_add_entry_with_term(term, replace=True)
        assert amidict.get_entry_count() == 1
        entry = amidict.get_lxml_entry(term)
        assert type(entry) is _Element

        assert term == entry.attrib[TERM]
        assert NAME not in entry.attrib

    def test_create_and_fail_on_add_entry_with_duplicate_term(self):
        term = "foo"
        amidict = AmiDictionary.create_minimal_dictionary()
        entry = amidict.create_and_add_entry_with_term(term)
        try:
            amidict.create_and_add_entry_with_term(term, replace=False)
            assert False, f"should fail with duplicate entry"
        except AMIDictError as e:
            assert True, "should raise duplicate error"

    def test_create_and_overwrite_duplicate_term(self):
        """
        setting term which already exists in dictionary, and replace with new term
        """
        term = "foo"
        amidict = AmiDictionary.create_minimal_dictionary()
        ami_entry = AmiEntry.create_from_element(amidict.create_and_add_entry_with_term(term))
        assert ami_entry.get_name() is None
        ami_entry.set_name("bar")
        assert ami_entry.get_name() == "bar"
        try:
            amidict.create_and_add_entry_with_term(term, replace=True)
            assert True, f"should overwrite duplicate entry"
        except AMIDictError as e:
            assert True, "should not raise duplicate error"

    # dictionary tests
    def test_minimal_dictionary(self):
        amidict = AmiDictionary.create_minimal_dictionary()
        assert amidict.get_version() is not None
        # amidict.check_validity()
        amidict.remove_attribute(VERSION)
        if amidict.get_version() is not None:
            raise AMIDictError("should have removed version")

        try:
            amidict.check_validity()
            raise AMIDictError("should fail is_valid()")
        except Exception as e:
            logging.info(f"failed test {e}")

    def test_get_duplicate_entries(self):
        """Dictionary has two entries for 'apical' but only one for 'cone'"""
        dup_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[DUPLICATE_ENTRIES])
        entries = dup_dict.get_lxml_entries()
        assert len(entries) == 4, "one duplicate term omitted"
        entries = dup_dict.find_entries_with_term("apical")
        assert entries is not None and len(entries) == 1
        entries = dup_dict.find_entries_with_term("zilch")
        assert entries is not None and len(entries) == 0

    def test_get_terms_from_valid_dictionary(self):
        """ETHNOBOT has no multiple entries'"""
        ethno_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        terms = ethno_dict.get_terms()
        assert terms is not None
        assert len(terms) == 8
        assert terms == ['anti-fumitory', 'adaptogen', 'homeopathy variable agent', 'ethnomedicinal agent',
                         'phytochemical agent', 'phytomedical agent', 'plant-extracted agent', 'lung-tonifying agent']

    def test_get_terms_from_invalid_dictionary(self):
        """DUPLICATE_ENTRIES has two entries for 'apical' and some missing terms"""
        dup_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[DUPLICATE_ENTRIES])
        terms = dup_dict.get_terms()
        assert terms == ['apical', 'flowering top', 'cone', 'pistil']

    # review dictionaries
    def test_mini_plant_part_is_valid(self):
        # pp_dict = AmiDictionary(setup_amidict[MINI_PLANT_PART])
        pp_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[MINI_PLANT_PART])
        if pp_dict is None:
            raise AMIDictError(f"test_dictionary_should_have_desc cannot read dictionary {pp_dict}")
        pp_dict.check_validity()

    def test_mini_mentha_tps_dict_is_valid(self):
        mentha_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[MINI_MENTHA])
        if mentha_dict is None:
            raise AMIDictError("cannot find/read mentha_dict")
        mentha_dict.check_validity()

    def test_ethnobot_dict_has_version(self):
        ethnobot_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        version = ethnobot_dict.get_version()
        assert version is not None
        assert AmiDictionary.is_valid_version_string(version)
        # assert ethnobot_dict.get_version() == "0.0.1"

    def test_ethnobot_dict_is_valid(self):
        logger.info(f" validating {ETHNOBOT_DICT}")
        ethnobot_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        ethnobot_dict.check_validity()
        # assert ethnobot_dict.get_version() == "0.0.1"

    def test_ethnobot_dict_has_8_entries(self):
        ethnobot_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        entries = ethnobot_dict.get_lxml_entries()
        assert len(entries) == 8

    def test_ethnobot_dict_entry_0_is_valid(self):
        ethnobot_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        entry0 = ethnobot_dict.get_lxml_entries()[0]
        AmiEntry.create_from_element(entry0).check_validity()

    def test_all_ethnobot_dict_entries_are_valid(self):
        ethnobot_dict = AmiDictionary.create_from_xml_file(self.create_file_dict()[ETHNOBOT_DICT])
        for entry in ethnobot_dict.get_lxml_entries():
            AmiEntry.create_from_element(entry).check_validity()

    # integrations

    def test_find_missing_wikidata_ids(self):
        ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
        lxml_entries = ami_dict.get_lxml_entries_with_missing_wikidata_ids()
        # missing_wikidata_ids = AmiEntry.get_wikidata_ids_for_entries(_entries)
        missing_wikidata_terms = AmiEntry.get_terms_for_lxml_entries(lxml_entries)
        assert missing_wikidata_terms == [

            'HFCs',
            'HCFCs',
            'CRF',
            'WMO',
            'NGHGI',
            'GWP',
            'FFI',
            'PBEs',
            'TCBA',
            'HCEs',
            'EBEs',
            'IBE',
            'RSD',
            'HDI',
            #            'CSP',
            'BECCS',
            'IAMs',
            # 'CDR',
            'ECR',
            'ETSs',
            #            'EVs',
            'ODSs',
            'HCS'
        ]

    def test_disambiguate_raw_wikidata_ids_in_dictionary(self):
        """
        find
        USABLE
        """
        ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
        _term_id_list = ami_dict.get_disambiguated_raw_wikidata_ids()
        assert len(_term_id_list) == 1
        assert _term_id_list[0] == ('GWP', ['Q901028'])

    #
    @unittest.skipUnless(VERYLONG, "lookup whole dictionaries")
    def test_lookup_missing_abbreviation_wikidata_ids_by_name(self):
        """
        scans dictionary for missing @wikidataID and searches wikidata by name/term
        USEFUL
        """
        ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
        lookup = ami_dict.lookup_missing_wikidata_ids()
        pprint.PrettyPrinter(indent=4).pprint(lookup.hits_dict)
        assert 15 <= len(lookup.hits_dict) <= 19
        """{   'Electric vehicles': {   'Q13629441': 'electric vehicle',
                             'Q17107666': 'Miles Electric Vehicles',
                             'Q1720353': 'Smith Electric Vehicles',
                             'Q5029961': 'Canadian Electric Vehicles',
                             'Q67371583': 'John Bradshaw Ltd.'},
    'Frequently Asked Questions': {   'Property:P9214': 'FAQ URL',
                                      'Q189293': 'FAQ',
                                      'Q76407407': 'Ffatrïoedd a busnesau a '
                                                   'gynorthwyir : cwestiynau '
                                                   'cyffredin = Supported '
                                                   'factories and businesses : '
                                                   'frequently asked '
                                                   'questions'},
    'Global Warming Potential': {'Property:P2565': 'global warming potential'},
    'Greenhouse Gas': {   'Q107315539': 'Greenhouse Gas Mitigation Workshop '
                                        '(2016)',
                          'Q167336': 'greenhouse gas',
                          'Q5604172': 'greenhouse gas emissions by the United '
                                      'Kingdom'},
"""

    @unittest.skipUnless(VERYLONG, "runs several chapters")
    def test_debug_chapter_dictionaries(self):
        self.debug_dict(dict_path=Resources.LOCAL_IPCC_CHAP07_ABB_DICT)
        self.debug_dict(dict_path=Resources.LOCAL_IPCC_CHAP07_MAN_DICT)
        self.debug_dict(dict_path=Resources.TEST_IPCC_CHAP08_ABB_DICT)
        self.debug_dict(dict_path=Resources.TEST_IPCC_CHAP08_MAN_DICT)

    @classmethod
    def debug_dict(cls, dict_path):
        logger.info(f"======={dict_path}=======")
        ami_dict = AmiDictionary.create_from_xml_file(dict_path)
        if ami_dict:
            lookup = ami_dict.lookup_missing_wikidata_ids()
            pprint.PrettyPrinter(indent=4).pprint(lookup.hits_dict)
        else:
            print(f"****Cannot find valid dict {dict_path}****")
            logging.error(f"Cannot find valid dict {dict_path}")

    @unittest.skipUnless(VERYLONG, "lookup whole dictionaries")
    def test_lookup_missing_manual_wikidata_ids_by_name(self):
        """
        scans dictionary for missing @wikidataID and searches wikidata by name/term
        USEFUL
        """
        ami_dict = AmiDictionary.create_from_xml_file(Resources.LOCAL_IPCC_CHAP07_MAN_DICT)
        lookup = ami_dict.lookup_missing_wikidata_ids()
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(lookup.hits_dict)
        assert 8 >= len(lookup.hits_dict) >= 6

    @unittest.skipUnless(VERYLONG, "lookup whole dictionaries")
    def test_lookup_missing_wikidata_ids_by_term(self):
        """
        scans dictionary for missing @wikidataID and searches wikidata by term
        USEFUL
        """
        ami_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
        lookup = ami_dict.lookup_missing_wikidata_ids(lookup_string=TERM)
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(lookup.hits_dict)
        assert 28 >= len(lookup.hits_dict) >= 22


    # TODO set changes over time
    def test_get_property_ids(self):
        """gets properties af a dictionary entry"""
        words = ["limonene"]
        dictionary, _ = AmiDictionary.create_dictionary_from_words(
            words, "test", "created from words",wikilangs=["en", "de"])
        dictionary.add_wikidata_from_terms()
        pprint.pprint(lxml.etree.tostring(dictionary.root).decode("UTF-8"))
        assert len(dictionary.entries) == 1
        wikidata_page = dictionary.create_wikidata_page(dictionary.entries[0])
        property_ids = wikidata_page.get_property_ids()
        assert len(property_ids) >= 60
        id_set = set(property_ids)
        test_set = {'P31', 'P279', 'P361', 'P2067', 'P274', 'P233', 'P2054', 'P2101', 'P2128', 'P2199'}
        assert test_set.issubset(id_set)


    @unittest.skipUnless(CEV_EXISTS, "requires local CEVOpen")
    def test_invasive(self):
        """

        """


        INVASIVE_DIR = Path(LOCAL_CEV_OPEN_DICT_DIR, "invasive_species")
        assert INVASIVE_DIR.exists(), f"INVASIVE_DIR should exists {INVASIVE_DIR}"
        dictionary_file = Path(INVASIVE_DIR, "invasive_plant.xml")
        assert dictionary_file.exists(), f"dictionary_file should exist {dictionary_file}"
        SPARQL_DIR = Path(INVASIVE_DIR, "sparql_output")
        assert SPARQL_DIR.exists(), f"SPARQL_DIR should exist {SPARQL_DIR}"
        rename_file = False

        sparql_files = glob.glob(os.path.join(SPARQL_DIR, "sparql_*.xml"))

        sparql_files.sort()
        sparql2amidict_dict = {
            "image": {
                "id_name": "item",
                "sparql_name": "image_link",
                "dict_name": "image",
            },
            "map": {
                "id_name": "item",
                "sparql_name": "taxon_range_map_image",
                "dict_name": "image",
            },
            # "taxon": {
            #     "id_name": "item",
            #     "sparql_name": "taxon",
            #     "dict_name": "synonym",
            # }
        }
        AmiDictionary.apply_dicts_and_sparql(dictionary_file, rename_file, sparql2amidict_dict, sparql_files)
        # TODO needs assert

    # LONG
    @unittest.skip("VERY LONG, SPARQL")
    @unittest.skipUnless(CEV_EXISTS, "requires local CEVOpen")
    def test_plant_genus(cls):
        """
        """

        DICT_DIR = Path(LOCAL_CEV_OPEN_DICT_DIR, "plant_genus")
        assert DICT_DIR.exists(), f"DICT_DIR should exist {DICT_DIR}"
        dictionary_file = Path(DICT_DIR, "plant_genus.xml")
        assert dictionary_file.exists(), f"dictionary_file should exists {dictionary_file}"
        SPARQL_DIR = Path(DICT_DIR, "raw")
        assert SPARQL_DIR.exists(), f"SPARQL_DIR should exist {SPARQL_DIR}"
        rename_file = False

        sparql_files = glob.glob(os.path.join(SPARQL_DIR, "sparql_test_concatenation.xml"))

        sparql_files.sort()
        sparql2amidict_dict = {
            "image": {
                "id_name": "plant_genus",
                "sparql_name": "images",
                "dict_name": "image",
            },
            "map": {
                "id_name": "plant_genus",
                "sparql_name": "taxon_range_map_image",
                "dict_name": "map",
            },
            "wikipedia": {
                "id_name": "plant_genus",
                "sparql_name": "wikipedia",
                "dict_name": "wikipedia",
            },
        }
        AmiDictionary.apply_dicts_and_sparql(dictionary_file, rename_file, sparql2amidict_dict, sparql_files)

    @unittest.skipUnless(CEV_EXISTS, "requires local CEVOpen")
    def test_compound(cls):
        """
        ???
        """

        DICT_DIR = Path(LOCAL_CEV_OPEN_DICT_DIR, "eoCompound")
        assert DICT_DIR.exists(), f"DICT_DIR should exists {DICT_DIR}"
        dictionary_file = Path(DICT_DIR, "plant_compound.xml")
        assert (os.path.exists(dictionary_file)), f"dictionary_file should exist {dictionary_file}"
        SPARQL_DIR = Path(DICT_DIR, "raw")
        SPARQL_DIR = DICT_DIR
        assert SPARQL_DIR.exists(), f"SPARQL_DIR should exist {SPARQL_DIR}"
        rename_file = False

        sparql_files = glob.glob(os.path.join(SPARQL_DIR, "sparql_6.xml"))
        sparql_files.sort()
        sparql2amidict_dict = {
            "image": {
                "id_name": "item",
                "sparql_name": "t",
                "dict_name": "image",
            },
            "chemform": {
                "id_name": "item",
                "sparql_name": "chemical_formula",
                "dict_name": "chemical_formula",
            },
            "wikipedia": {
                "id_name": "plant_genus",
                "sparql_name": "wikipedia",
                "dict_name": "wikipedia",
            },
            # "taxon": {
            #     "id_name": "item",
            #     "sparql_name": "taxon",
            #     "dict_name": "taxon",
            # }
        }

        AmiDictionary.apply_dicts_and_sparql(dictionary_file, rename_file, sparql2amidict_dict, sparql_files)

    @unittest.skipUnless(CEV_EXISTS, "requires local CEVOpen")
    def test_plant_part_sparql(cls):
        """
        Takes WD-SPARQL-XML output (sparql.xml) and maps to AMIDictionary (eo_plant_part.xml)

        """
        # current dictionary does not need updating

        logger.debug(f"***test_plant_part")
        DICT_DIR = Path(LOCAL_CEV_OPEN_DICT_DIR, "eoPlantPart")
        DICT_DIR = Path(TEST_RESOURCE_DIR, "eoPlantPart")
        assert DICT_DIR.exists(), f"{DICT_DIR} should exist"
        dictionary_file = Path(DICT_DIR, "eoplant_part.xml")
        assert dictionary_file.exists(), f"dictionary_file should exist {dictionary_file}"
        SPARQL_DIR = Path(DICT_DIR, "raw")
        SPARQL_DIR = DICT_DIR
        assert SPARQL_DIR.exists(), f"SPARQL_DIR should exists {SPARQL_DIR}"
        rename_file = False

        sparql_files = glob.glob(os.path.join(SPARQL_DIR, "sparql.xml"))

        sparql_files.sort()
        sparql2amidict_dict = {
            "image": {
                "id_name": "item",
                "sparql_name": "image",
                "dict_name": "image",
            },
        }

        AmiDictionary.apply_dicts_and_sparql(dictionary_file, rename_file, sparql2amidict_dict, sparql_files)

    def test_merge_dicts_ipcc_same_chap(self):
        """test merge dictionaries from IPCC (heavy commonality)"""

        abb2_dict = AmiDictionary.create_from_xml_file(Resources.TEST_IPCC_CHAP02_ABB_DICT)
        abb2_set = abb2_dict.get_or_create_term_set()
        assert abb2_set == {
            'BECCS',
            #            'CBEs',
            #            'CDR',
            'CRF',
            #            'CSP',
            'EBEs', 'ECR',
            # 'EET',
            'ETSs',
            'EU ETS',
            #            'EVs',
            #            'F-gases', 'FAQs',
            'FFI',
            # 'GDP', 'GHG',
            'GTP',
            'GWP',
            #            'GWP100',
            'HCEs', 'HCFCs',
            'HCS', 'HDI', 'HFCs', 'IAMs',
            'IBE', 'LULUCF', 'NGHGI', 'ODSs',
            'PBEs',
            # 'PFCs',
            'RGGI', 'RSD',
            'TCBA',
            # 'UNFCCC',
            'WMO'
        }, f"abb2 set {abb2_set}"

        man2_dict = AmiDictionary.create_from_xml_file(
            Path(Resources.TEST_IPCC_CHAP02_DICT, "ip_3_2_emissions_man.xml"))
        man2_set = man2_dict.get_or_create_term_set()
        assert man2_set == {
            'CAIT', 'CEDS', 'CGTP', 'CO2-equivalent emission',
            'CRF', 'EDGAR', 'FAOSTAT', 'FFI', 'FOLU', 'Final Energy Demand', 'GTP', 'GWP',
            'GWP100', 'GtCO2eq', 'LULUCF', 'NMVOC',
            'PRIMAP', 'Paris Agreement', 'Primary Energy', 'Primary Energy Conversion',
            'SLCF', 'SRES', 'SSP', 'UNFCCC', 'WMO',
            'atmospheric lifetime', 'baseline scenario',
            'carbon budget', 'carbon pricing',
            'cumulative CO2 emissions', 'demand side solutions',
            'emission inventory', 'emission sectors',
            'emissions factor', 'emissions trajectory',
            'fluorinated gas', 'social discount rate',
            'top down atmospheric measurement'
        }, f"man2 set {man2_set}"

        # phrases
        phr2_dict = AmiDictionary.create_from_xml_file(
            Path(Resources.TEST_IPCC_CHAP02_DICT, "ip_3_2_emissions_phr.xml"))
        phr2_set = phr2_dict.get_or_create_term_set()
        assert phr2_set == {
            'BECCS', 'CBEs', 'CDR', 'CRF', 'CSP', 'EBEs', 'ECR', 'EET',
            'ETSs', 'EU ETS', 'EVs', 'F-gases',
            'FAQs', 'FFI', 'GDP', 'GHG', 'GTP', 'GWP', 'GWP-100', 'GWP100',
            'HCEs', 'HCFCs', 'HCS', 'HDI', 'HFCs', 'IAMs', 'IBE', 'LULUCF',
            'NGHGI', 'ODSs', 'PBEs', 'PFCs', 'RGGI', 'RSD', 'TCBA', 'UNFCCC', 'WMO'
        }

        # terms common to abbrev and manual
        abb_man_set = abb2_set.intersection(man2_set)
        assert len(abb_man_set) == 6, f"man2 set {len(abb_man_set)}"
        assert abb_man_set == {
            'GTP', 'FFI',
            #            'GWP', 'UNFCCC',
            'WMO', 'GWP', 'CRF', 'LULUCF'}


    def test_entries_have_ids(self):
        dict_file = Path(Resources.TEST_IPCC_DICT_DIR, "raw_linked_dict.xml")
        assert dict_file.exists(), f"file {dict_file} should exist"
        linked_dict = AmiDictionary.create_from_xml_file(dict_file)
        assert linked_dict is not None
        assert linked_dict.get_entry_count() == 17, f"found {linked_dict.get_entry_count()} entries"
        # assert linked_dict.get_lxml_entries_with_ids() == 2

    def test_entry_id(self):
        """
        Tests whether entries have ids, and (later) creates default values
        """
        dict_dir = Path(Resources.TEST_RESOURCES_DIR, "eoCompound")
        assert dict_dir.exists(), f"{dict_dir} should exist"
        ambig_dict_file = Path(dict_dir, "disambig.xml")
        assert ambig_dict_file.exists(), f"{ambig_dict_file} should exist"
        disambig_dict = AmiDictionary.create_from_xml_file(ambig_dict_file)
        assert disambig_dict is not None

        assert disambig_dict.get_entry_count() == 9, f"dictionary should have 9 entries"
        assert disambig_dict.get_entry_ids() == ['____alloaromadendrene']

    def test_dict_commands_COMMAND_OK(self):
        """
        prints help
        """
        amilib = AmiLib()
        amilib.run_command(["--help"])

        amilib.run_command(["DICT", "--help"])


    def test_search_with_dictionary_and_make_links_IMPORTANT(self):
        """
        uses a simple dictionary to search WG chapter (wg2/ch03) *html_with_ids)

        Returns
        -------


        """

        chapter_file = Path(Resources.TEST_RESOURCES_DIR, "ipcc", "wg3", "Chapter03", f"{self.HTML_WITH_IDS}.html")
        paras = HtmlLib._extract_paras_with_ids(chapter_file, count=1163)
        xml_dict_path = Path(Resources.TEST_RESOURCES_DIR, "dictionary", "climate", "climate_words.xml")
        dictionary = AmiDictionary.create_from_xml_file(xml_dict_path)
        assert dictionary is not None
        phrases = dictionary.get_terms()
        html_path = Path(Resources.TEST_RESOURCES_DIR, "dictionary", "climate", "climate_words.html")
        dictionary.create_html_write_to_file(html_path, debug=True)
        dictionary.location = html_path
        assert len(phrases) == 11
        para_phrase_dict = HtmlLib.search_phrases_in_paragraphs(paras, phrases, markup=html_path)
        chapter_elem = paras[0].xpath("/html")[0]
        chapter_outpath = Path(Resources.TEMP_DIR, "ipcc", "Chapter03", "marked_up.html", debug=True)
        HtmlLib.write_html_file(chapter_elem, chapter_outpath, debug=True)

    def test_search_with_dictionary_and_make_links_WORKFLOW(self):
        """
        uses a simple dictionary to search WG chapter (wg2/ch03) *html_with_ids)

        Returns
        -------


        """
        stem = "carbon_cycle"
        words_path = Path(Resources.TEST_RESOURCES_DIR, "wordlists", f"{stem}_edited.txt")
        assert words_path.exists()

        chapter_file = Path(Resources.TEST_RESOURCES_DIR, "ar6", "wg1", "Chapter05", f"{self.HTML_WITH_IDS}.html")
        assert chapter_file.exists()

        paras = HtmlLib._extract_paras_with_ids(chapter_file, count=1724)

        dictionary, outpath = AmiDictionary.create_dictionary_from_wordfile(words_path)
        assert dictionary is not None
        assert len(dictionary.get_terms()) == 43

        xml_dict_path = Path(Resources.TEMP_DIR, "dictionary", "climate", f"{stem}.xml")
        dictionary.create_html_write_to_file(xml_dict_path, debug=True)
        assert xml_dict_path.exists()

        html_dict_path = Path(Resources.TEMP_DIR, "dictionary", "climate", f"{stem}.html")
        dictionary.create_html_write_to_file(html_dict_path, debug=True)
        assert html_dict_path.exists()

        phrases = dictionary.get_terms()
        dictionary.location = html_dict_path
        assert len(phrases) == 43
        para_phrase_dict = HtmlLib.search_phrases_in_paragraphs(paras, phrases, markup=html_dict_path)

        # write marked_up html
        chapter_elem = paras[0].xpath("/html")[0]
        chapter_outpath = Path(Resources.TEMP_DIR, "ipcc", "wg1", "Chapter05", "marked_up.html", debug=True)
        HtmlLib.write_html_file(chapter_elem, chapter_outpath, debug=True)
        assert  chapter_outpath.exists()

    def test_search_with_dictionary_and_make_links_code(self):
        """
        uses a simple dictionary to search WG chapter (wg2/ch03) *html_with_ids) and add hyperlinks
        -------

        """
        stem = "carbon_cycle"
        chapter_file = Path(Resources.TEST_RESOURCES_DIR, "ar6", "wg1", "Chapter05", f"{self.HTML_WITH_IDS}.html")
        chapter_outpath = Path(Resources.TEMP_DIR, "ipcc", "wg1", "Chapter05", "marked_up.html", debug=True)
        FileLib.delete_file(chapter_outpath)
        html_dict_path = Path(Resources.TEMP_DIR, "dictionary", "climate", f"{stem}.html")

        AmiDictionary.read_html_dictionary_and_markup_html_file(
            chapter_file, chapter_outpath, html_dict_path=html_dict_path)
        assert chapter_outpath.exists()

    def test_search_with_dictionary_and_make_links_COMMANDLINE(self):
        """
        same logic and files as test_search_with_dictionary_and_make_links_CODE. Check that that runs
        """

        stem = "carbon_cycle"
        chapter_file = str(Path(Resources.TEST_RESOURCES_DIR, "ar6", "wg1", "Chapter05", f"{self.HTML_WITH_IDS}.html"))
        chapter_outpath = str(Path(Resources.TEMP_DIR, "ipcc", "wg1", "Chapter05", "marked_up.html", debug=True))
        FileLib.delete_file(chapter_outpath)
        html_dict_path = str(Path(Resources.TEMP_DIR, "dictionary", "climate", f"{stem}.html"))

        # commandline
        args = [
            "DICT",
            "--inpath", chapter_file,
            "--outpath", chapter_outpath,
            "--dict", html_dict_path,
            "--operation", MARKUP_FILE,
        ]
        logger.info(f"cmd> {' '.join(args)}")
        AmiLib().run_command(args)
        assert Path(chapter_outpath).exists()

    def test_args_help(self):
        args = [
            "DICT",
            "--help"
        ]
        AmiLib().run_command(args)

class AmiEntryTest(AmiAnyTest):
    """
    test functionality of AmiEntry
    """

    def dummy(self):
        pass


class ValidateTest(AmiAnyTest):
    """
    test validity of AmiEntry and AmiDictionary
    """

    def test_read_wellformed_dictionary(self):
        """test can create from XML string
        includes well-formed and non-well-formed XML
        """
        dict_str = """
        <dictionary title='foo'>
        </dictionary>
        """
        ami_dict = AmiDictionary.create_dictionary_from_xml_string(dict_str)
        assert ami_dict is not None

        assert ami_dict.root.tag == "dictionary"

        dict_str = """
        <diktionary title='foo'>
        </dictionary>
        """
        try:
            ami_dict = AmiDictionary.create_dictionary_from_xml_string(dict_str)
        except XMLSyntaxError as e:
            logger.warning(f"xml error {e}")

    def test_one_entry_dict_is_ami_dictionary(self):
        """require the attribute to be present but does not check value"""
        setup_dict = AmiDictionaryTest().create_file_dict()
        one_dict = setup_dict[ONE_ENTRY_DICT]
        assert type(one_dict) is AmiDictionary, f"fila is not AmiDictionary {one_dict}"

    def test_dict1_has_version_attribute(self):
        """require the version attribute to be present but does not check value"""
        setup_dict = AmiDictionaryTest().create_file_dict()
        one_dict = setup_dict[DICTFILE1]
        amidict = AmiDictionary.create_from_xml_file(Path(one_dict))
        version = amidict.get_version()
        assert version == STARTING_VERSION

    def test_dict1_with_missing_version_attribute_is_not_valid(self):
        """
        require the version attribute to have starting value
        """
        setup_dict = AmiDictionaryTest().create_file_dict()
        amidict = AmiDictionary.create_from_xml_file(Path(setup_dict[DICTFILE1]))
        version = amidict.get_version()
        assert version == STARTING_VERSION

    def test_one_entry_dict_has_version_attribute(self):
        """require the attribiute to be present but does not check value"""
        setup_dict = AmiDictionaryTest().create_file_dict()
        one_dict = setup_dict[ONE_ENTRY_DICT]
        assert one_dict is not None
        version = one_dict.get_version()
        assert version == "1.2.3"

    def test_dictionary_has_version(self):
        """require the attribute to be present but does not check value"""
        setup_dict = AmiDictionaryTest().create_file_dict()
        one_dict = setup_dict[ONE_ENTRY_DICT]
        version = one_dict.get_version()
        assert version is not None, "missing version"

    @unittest.skip("superseded by Validator")
    def test_dictionary_has_valid_version(self):
        """require the attribute to be present but does not check value"""
        setup_dict = self.setup()
        one_dict = setup_dict[ONE_ENTRY_DICT]
        validator = AmiDictValidator(one_dict)
        version = one_dict.get_version()
        # assert validator. f"invalid version {version}"

    def test_catch_invalid_version(self):
        minimal_dict = AmiDictionary.create_minimal_dictionary()
        try:
            minimal_dict.set_version("1.2.a")
            raise AMIDictError("should catch bad version error")
        except AMIDictError as e:
            """should catch bad version"""
            # logger.warning(f"caught expected error")

    def test_dict_has_xml_title(self):
        """has root dictionary element got title attribute?
        e.g. <dictionary title='dict1'> ..."""
        setup_dict = AmiDictionaryTest().create_file_dict()
        root = setup_dict[ROOT]
        assert root.attrib[TITLE] == "dict1"

    def test_dict_title_matches_filename(self):
        setup_dict = AmiDictionaryTest().create_file_dict()
        root = setup_dict[ROOT]
        last_path = setup_dict[DICTFILE1].stem
        logger.debug(last_path)
        assert root.attrib["title"] == last_path

    def test_validate_dictionary_created_from_words(self):
        """
        validate dictionary created from list of words
        """
        ami_dictionary, _ = AmiDictionary.create_dictionary_from_words(["curlicue", "bread"],
                                                                    title="test",
                                                                    desc="test dictionary")
        validator = AmiDictValidator(ami_dictionary)

    def test_validate_dictionary_created_from_wordfile(self):
        """
        validate dictionary created from list of words
        """
        file = Path(Resources.TEST_RESOURCES_DIR, "wordlists", "carbon_cycle.txt")
        ami_dictionary, _ = AmiDictionary.create_dictionary_from_wordfile(file,
                                                                    title="test",
                                                                    desc="test dictionary")
        validator = AmiDictValidator(ami_dictionary)


    def test_xml_dictionary(self):
        """
        reads previously created XML dictionary
        """
        file = Path(Resources.TEST_RESOURCES_DIR, "wordlists", "xml", "climate_words_dict.xml")
        ami_dictionary = AmiDictionary.create_from_xml_file(file)
        validator = AmiDictValidator(ami_dictionary)


class DictionaryCreationTest(AmiAnyTest):
    """
    test methods which create dictionaries
    (from lists of strings, files, urls, sparql etc.

    """

    def test_read_wellformed_dictionary(self):
        """test can create from XML string
        includes well-formed and non-well-formed XML
        """
        dict_str = """
        <dictionary title='foo'>
        </dictionary>
        """
        ami_dict = AmiDictionary.create_dictionary_from_xml_string(dict_str)
        assert ami_dict is not None

        assert ami_dict.root.tag == "dictionary"

        dict_str = """
        <diktionary title='foo'>
        </dictionary>
        """
        try:
            ami_dict = AmiDictionary.create_dictionary_from_xml_string(dict_str)
        except XMLSyntaxError as e:
            logger.warning(f"xml error {e}")


    def test_dictionary_creation(self):
        """
        unit test
        """
        amidict = AmiDictionary.create_minimal_dictionary()
        assert amidict is not None
        assert amidict.get_version() == STARTING_VERSION


    def test_create_dictionary_from_list_of_strings_in_file(dictionary
#                                                                         , words_txt, entry_count
                                                                         ):
        """
        read file with lists of strings and create AmiDictionary
        """
        dictionary = "climate"
        words_txt = "ar5_wg3_food_security_words.txt"
        entry_count = 60
        dictionary_in_dir = Path(Resources.TEST_RESOURCES_DIR, "dictionary")
        out_dict_dir = Path(Resources.TEMP_DIR, "dictionary", "climate")
        words_file = Path(dictionary_in_dir, dictionary, words_txt)
        assert words_file.exists(), f"{words_file} should exist"
# create dictionary without save or lookup
        words_dictionary, _ = AmiDictionary.create_dictionary_from_wordfile(wordfile=words_file)
        assert words_dictionary is not None, (f"words dictionary from {words_file} must not be None")
        assert words_dictionary.get_entry_count() == entry_count, f"should have {entry_count} entries , found {words_dictionary.get_entry_count()}"

# parametrize doesn't work easily withe unittest subclasses
    # @pytest.mark.parametrize("dictionary,words_txt,entry_count",
    #                          [("climate", "ar5_wg3_food_security_words.txt", 60)])
    # def test_create_dictionary_from_list_of_strings_in_file_parameterize(dictionary, words_txt, entry_count):
    #     """
    #     read file with lists of strings and create AmiDictionary
    #     """
    #     dictionary_in_dir = Path(Resources.TEST_RESOURCES_DIR, "dictionary")
    #     out_dict_dir = Path(Resources.TEMP_DIR, "dictionary", "climate")
    #     words_file = Path(dictionary_in_dir, dictionary, words_txt)
    #     assert words_file.exists(), f"{words_file} should exist"
    #     # create dictionary without save or lookup
    #     words_dictionary, _ = AmiDictionary.create_dictionary_from_wordfile(wordfile=words_file)
    #     assert words_dictionary is not None, (f"words dictionary from {words_file} must not be None")
    #
    #     assert words_dictionary.get_entry_count() == entry_count, f"should have {entry_count} entries , found {words_dictionary.get_entry_count()}"

    def test_createdictionary_from_list_of_strings_in_file_and_lookup_wikidata(self):
# add title, description, and save
        out_dict_dir = Path(Resources.TEMP_DIR, "dictionary", "climate")
        words_file = Path(Resources.TEST_RESOURCES_DIR, "dictionary", "climate",
                          "ar5_wg3_food_security_words.txt")
        max_entries = 2 # to reduce time

        description = "food security terms in AR5_WG3_ch_5 selected by Anmol Negi"
        # add title, description, and save
        description = "food security terms in AR5_WG3_ch_5 selected by Anmol Negi and wikidata lookup"
        words_dictionary, outfile = AmiDictionary.create_dictionary_from_wordfile(
            wordfile=words_file, desc=description, title="food_security", outdir=out_dict_dir, wikidata=True,
            max_entries=max_entries, debug=True)
        assert words_dictionary is not None, f"words dictionary from {words_file} must not be None"
        assert words_dictionary.get_entry_count() == max_entries, \
            f"should have {max_entries} entries , found {words_dictionary.get_entry_count()}"
        assert outfile.exists(), f"should create {outfile}"


    def test_process_args_build_dictionary_add_wikidata_CMD_OK(self):
        """commandline
        reads list of words, creates a dictionary, looks up Wikidata
        Wikidata has <entry>s in in format:
    <entry name="UNFCCC" term="UNFCCC">
    <div qid="Q21707860">
        <a href="wikidataURL/Q21707860">Wikidata</a>
        <p role="title">Paris Agreement</p>
        <p role="description">international agreement from 12 December 2015 within the United Nations Framework Convention on Climate Change (UNFCCC) dealing with greenhouse gas emissions mitigation, adaptation and finance starting in the year 2020</p>
    </div>
    <div qid="Q208645">
        <a href="wikidataURL/Q208645">Wikidata</a>
        <p role="title">United Nations Framework Convention on Climate Change</p>
        <p role="description">international treaty</p>
    </div>
    ...

        Note that wikidata lookup does not necessarily put the best answers first
        """
        amilib = AmiLib()
        title = "small_2"
        title_path = Path(Resources.TEST_RESOURCES_DIR, "wordlists", f"{title}.txt")
        assert title_path.exists(), f"{title_path} should exist"
        title_txt = f"{title_path}"
        logger.debug(f"words {title_txt}")
        with open(title_txt, "r") as f:
            logger.debug(f"words ==> {f.readlines()}")
        title_dict = Path(Resources.TEMP_DIR, "words", f"{title}_dict.xml")
        FileLib.delete_file(title_dict)

        word_xml = f"{Path(Resources.TEMP_DIR, '{title}.xml')}"
        args = ["DICT",
                "--words", f"{title_path}",
                "--dict", f"{title_dict}",
                "--wikidata"
                ]
        logger.debug(f" s2 {args}")
        amilib.run_command(args)
        assert Path(title_dict).exists(), f"should write dictionary {title_dict}"
        print(f"wrote {title_dict}")

    # integrations
    def test_create_dictionary_from_list_of_string(self):
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        title = "foobar"
        directory = None
        amidict, _ = AmiDictionary.create_dictionary_from_words(terms, title)
        assert amidict is not None
        title = amidict.root.attrib[TITLE]
        assert title == "foobar"

    def test_create_dictionary_from_list_of_string_and_save(self):
        """
        Create dictionary from list of strings (no lookup)
        """
        terms = ["acetone", "benzene", "chloroform", "DMSO", "ethanol"]
        title = "solvents"
        temp_dir = Path(AmiAnyTest.TEMP_DIR, "dictionary")
        temp_dir.mkdir(exist_ok=True, parents=True)
        amidict, dictfile = AmiDictionary.create_dictionary_from_words(terms, title=title, outdir=temp_dir)
        outfile = Path(temp_dir, title + ".xml")
        assert Path(dictfile) == Path(temp_dir, title + ".xml") and os.path.exists(dictfile)
        assert amidict.get_entry_count() == 5

    def test_make_dict_from_file_CMD_OK(self):

        amilib = AmiLib()
        words_file = Path(TEST_RESOURCE_DIR, "wordlists", "climate_words.txt")
        dictfile = Path(Resources.TEMP_DIR, "words", "climate_words.xml")
        expected_count = 11
        FileLib.delete_file(dictfile)
        amilib.run_command(["DICT",
                            "--words", words_file,
                            "--dict", dictfile
                            ])
        assert Path(dictfile).exists()
        ami_dictionary = AmiDictionary.create_from_xml_file(dictfile)
        assert (c := ami_dictionary.get_entry_count()) == expected_count, \
            f"dictionary should contain {expected_count} found {c}"

    def test_make_dict_from_words_CMD_OK(self):
        amilib = AmiLib()
        words_file = Path(TEST_RESOURCE_DIR, "wordlists", "climate_words.txt")
        dictfile = Path(Resources.TEMP_DIR, "words", "climate_words.xml")
        expected_count = 1
        FileLib.delete_file(dictfile)
        amilib.run_command(["DICT",
                            "--words", "bread",
                            "--dict", dictfile
                            ])
        assert Path(dictfile).exists()
        ami_dictionary = AmiDictionary.create_from_xml_file(dictfile)
        assert (c := ami_dictionary.get_entry_count()) == expected_count, \
            f"dictionary should contain {expected_count} found {c}"

    def test_make_dict_from_words(self):
        """
        filas on "kangaroo"
        Reason unknown
        """
        amilib = AmiLib()
        dictfile = Path(Resources.TEMP_DIR, "words", "kangaroo.xml")
        expected_count = 1
        FileLib.delete_file(dictfile)
        amilib.run_command(["DICT",
                            "--words", "kangaroo",
                            "--dict", dictfile,
                            ])
        assert Path(dictfile).exists()
        ami_dictionary = AmiDictionary.create_from_xml_file(dictfile)
        assert (c := ami_dictionary.get_entry_count()) >= expected_count, \
            f"dictionary should contain {expected_count} found {c}"


    def test_create_dictionary_from_csv(self):
        """This does lookup, unfortunately Wikidata lookup changes as terms are added so this test is a mess"""
        csv_term_file = Path(Resources.TEST_IPCC_CHAP08_DICT, "urban_terms_1.csv")
        assert os.path.exists(str(csv_term_file)), f"csv_term_file should exist {csv_term_file}"
        keyword_dict = AmiDictionary.create_dictionary_from_csv(csv_term_file, col_name='keyword/phrase',
                                                                title="urban_terms_1")
        print(f"keyword_dict {keyword_dict}")
        entry = keyword_dict.get_first_ami_entry()
        assert entry.get_term() == "urbanization"
        assert entry.get_wikidata_id() == "Q161078" or entry.get_wikidata_id() == 'Q1508'
        assert len(keyword_dict.entries) == 1
        root_elem = keyword_dict.root
        assert root_elem.tag == 'dictionary'
        entry0 = root_elem.xpath("entry")[0]
        assert entry0.attrib["desc"] == 'highly urbanized city in Metro Manila, Philippines'
        # assert entry0.attrib["desc"] == 'longterm population movements (shift) from rural to urban areas'
        assert entry0.attrib["name"] == 'urbanization'
        # assert str(root) == \
        #        '<dictionary title="urban_terms_1"><entry name="urbanization" term="urbanization" wikidataID="Q161078" wikidataURL="https://www.wikidata.org/entity/Q161078" desc="longterm population movements (shift) from rural to urban areas;gradual increase in the proportion of people living in urban areas, and the ways in which each society adapts to the change;process by which towns and cities are formed and become larger"><wikidataHit type="wikidata_hits">Q5381005</wikidataHit><wikidataHit type="wikidata_hits">Q23580084</wikidataHit><wikidataHit type="wikidata_hits">Q2608153</wikidataHit><wikidataHit type="wikidata_hits">Q95443723</wikidataHit></entry></dictionary>'

    def test_validate_linked_dict(self):
        """
        tests a veriety of exploratory concepts such as inter- and intra-dictionary links
        DEVELOP 2022-12
        """
        dict_path = Path(Resources.TEST_IPCC_DICT_DIR, "raw_linked_dict.xml")
        assert dict_path.exists(), f"{dict_path} should exist"
        dictx = AmiDictionary.create_from_xml_file(dict_path)
        assert dictx is not None
        assert dictx.get_entry_count() == 17, f"dictionary has {dictx.get_entry_count()} entries"
        assert dictx.get_version() == '0.0.5', f"dictionary has version {dictx.get_version()}"
        terms = dictx.get_terms()
        assert len(terms) > 0, f"found {len(terms)}"
        assert terms == ['carbon budget',
                         'ffi',
                         'folu',
                         'fluorinated gas',
                         'carbon pricing',
                         'edgar',
                         'cait',
                         'ceds',
                         'crf',
                         'emissions factor',
                         'emission inventory',
                         'emission sectors',
                         'gwp100',
                         'slcf',
                         'cgtp',
                         'co2-equivalent emission',
                         'baseline scenario'], f'found {terms}'
    def test_create_dictionary_from_sparql(self):
        """
        read spqrql file and create dictionary
        then update from another sparql file
        """
        # PLANT = os.path.join(PHYSCHEM_RESOURCES, "plant")
        PLANT = os.path.join(Resources.TEST_RESOURCES_DIR, "plant")
        sparql_file = os.path.join(PLANT, "plant_part_sparql.xml")
        dictionary_file = os.path.join(PLANT, "eoplant_part.xml")
        assert Path(sparql_file).exists(), f"sparql_file should exist {sparql_file}"
        assert Path(dictionary_file).exists(), f"dictionary_file should exist {dictionary_file}"
        """
        <result>
            <binding name='item'>
                <uri>http://www.wikidata.org/entity/Q2923673</uri>
            </binding>
            <binding name='image'>
                <uri>http://commons.wikimedia.org/wiki/Special:FilePath/White%20Branches.jpg</uri>
            </binding>
        </result>
"""
        sparql_to_dictionary = {
            "id_name": "item",
            "sparql_name": "image",
            "dict_name": "image",
        }
        dictionary = AmiDictionary.create_from_xml_file(dictionary_file)
        wikidata_sparql = WikidataSparql(dictionary)
        wikidata_sparql.update_from_sparql(sparql_file, sparql_to_dictionary)
        outdir = Path(AmiAnyTest.TEMP_DIR, "sparql")
        outdir.mkdir(exist_ok=True)
        dictionary.write_to_dir(outdir)
        outpath = Path(outdir, "eoplant_part.xml")
        assert outpath.exists(), f"{outpath} should have been written"

    @unittest.skipUnless(False, "LONG DOWNLOAD")
    def test_create_dictionary_terpenes(self):
        words = ["limonene", "alpha-pinene", "Lantana camara"]
        description = "created from words"
        title = "test"
        dictionary, _ = AmiDictionary.create_dictionary_from_words(words, title=title, desc=description, wikidata=True)
        assert len(dictionary.entries) == 3
        first_entry = dictionary.get_first_ami_entry()
        term = first_entry.term
        assert term == "limonene", f"first term is {term}"

    # @unittest.skip("LONG")
    def test_create_dictionary_from_list_of_string_and_add_wikidata(self):
        terms = ["acetone",
                 "chloroform",
                 # "DMSO",
                 # "ethanol"
                 ]
        amidict, _ = AmiDictionary.create_dictionary_from_words(terms, title="solvents", wikidata=True)
        temp_dir = Path(AmiAnyTest.TEMP_DIR, "dictionary")
        dictfile = amidict.write_to_dir(temp_dir)

        with open(dictfile, "r") as f:
            dict_text = f.read()
        dict_text = re.sub("date=\"[^\"]*\"", "date=\"TODAY\"", dict_text)

        # note, the date is stripped as it changes with each run
        text1 = """<dictionary version="0.0.1" title="solvents" encoding="UTF-8">
      <metadata user="pm286" date="TODAY"/>
      <entry term="acetone" wikidataID="Q49546" description="chemical compound"/>
      <entry term="chloroform" wikidataID="Q172275" description="chemical compound"/>
    </dictionary>
    """
        # assert text1 == dict_text, f"{text1} != {dict_text}"
        # TODO remove user from metadata

    def test_create_dictionary_from_list_of_string(self):
        terms = ["foo", "bar", "plugh", "xyzzy", "baz"]
        title = "foobar"
        directory = None
        amidict, _ = AmiDictionary.create_dictionary_from_words(terms, title)
        assert amidict is not None
        title = amidict.root.attrib[TITLE]
        assert title == "foobar"

    def test_create_dictionary_from_list_of_string_and_save(self):
        """
        Create dictionary from list of strings (no lookup)
        """
        terms = ["acetone", "benzene", "chloroform", "DMSO", "ethanol"]
        title = "solvents"
        temp_dir = Path(AmiAnyTest.TEMP_DIR, "dictionary")
        temp_dir.mkdir(exist_ok=True, parents=True)
        amidict, dictfile = AmiDictionary.create_dictionary_from_words(terms, title=title, outdir=temp_dir)
        outfile = Path(temp_dir, title + ".xml")
        assert Path(dictfile) == Path(temp_dir, title + ".xml") and os.path.exists(dictfile)
        assert amidict.get_entry_count() == 5

    def test_create_dictionary_from_list_of_string_save_and_compare(self):
        terms = ["acetone", "benzene", "chloroform", "DMSO", "ethanol"]
        temp_dir = Path(AmiAnyTest.TEMP_DIR, "dictionary")
        amidict, dictfile = AmiDictionary.create_dictionary_from_words(terms, title="solvents", outdir=temp_dir)
        with open(dictfile, "r") as f:
            dict_text = f.read()
        dict_text = re.sub("date=\"[^\"]*\"", "date=\"TODAY\"", dict_text)
        assert len(dict_text) > 200, "lines of dict_text"
        assert type(dict_text) is str, f"{type(dict_text)}"
        # note, the date is nstripped as it changes with each run
        text1 = """<dictionary version="0.0.1" title="solvents" encoding="UTF-8">
      <metadata user="pm286" date="TODAY"/>
      <entry term="acetone"/>
      <entry term="benzene"/>
      <entry term="chloroform"/>
      <entry term="DMSO"/>
      <entry term="ethanol"/>
    </dictionary>
    """
        # assert text1 == dict_text, f"{text1} != {dict_text}"
        # TODO remove user from metadata

    def test_create_dictionary_from_url(self):
        """lookup entry in Github repository
        note: depends on INTERNET"""
        mentha_url = "https://raw.githubusercontent.com/petermr/pyami/main/py4ami/resources/amidicts/mentha_tps.xml"
        mentha_dict = AmiDictionary.create_dictionary_from_url(mentha_url)
        assert len(mentha_dict.get_lxml_entries()) == 1
        assert mentha_dict.get_first_ami_entry().get_term() == "1,8-cineole synthase"
        assert mentha_dict.get_version() == "0.0.3"

    def test_make_dict_from_words_and_add_wikimedia_image(self):
        """
        Reason unknown
        """
        amilib = AmiLib()
        dictfile = Path(Resources.TEMP_DIR, "words", "kangaroo.xml")
        expected_count = 1
        FileLib.delete_file(dictfile)
        amilib.run_command(["DICT",
                            "--words", "kangaroo",
                            "--dict", dictfile,
                            ])
        assert Path(dictfile).exists()
        ami_dictionary = AmiDictionary.create_from_xml_file(dictfile)
        assert (c := ami_dictionary.get_entry_count()) >= expected_count, \
            f"dictionary should contain {expected_count} found {c}"

    def test_make_dict_from_wordlist_add_figures(self):
        stems = [
            # "breward",
            "carbon_cycle_noabb",
            # "chap2",
            # "chap5",
            # "chap6_7_8",
            # "food_ecosystem",
            # "human_influence",
            # "inews15",
            # "non_wikipedia",
            # "poverty",
            # "small_2",
            # "water_cyclone"
            ""

            ]

        for stem in stems:
            logger.info(f"creating figure entries for {stem}")
            self.create_dict_for_word_file(stem, wiktionary=True)

    def test_more_words(self):
        pass

    def create_dict_for_word_file(self, stem, wiktionary=True, input_file=None, output_dict=None):
        """
        Reads a wordlist file and creates output dictionary with figures
        :param stem: stem of files, also used as title
        :param input_file: if None, defaults to TEST_RESOURCES_DIR, "wordlists", f"{stem}.txt"
        :param output_dict: if None, defaults to TEMP_DIR, "words", "html", f"{stem}.html"

        """
        if stem is None or stem.strip() == "":
            logger.error("no title given for dictionary")
            return None
        if input_file is None:
            input_file = Path(Resources.TEST_RESOURCES_DIR, "wordlists", f"{stem}.txt")
        if output_dict is None:
            output_dict = Path(Resources.TEMP_DIR, "words", "html", f"{stem}.html")


        args = [
            "DICT",
            "--words", input_file,
            "--dict", output_dict,
            "--figures", "wikipedia",
            "--title", stem,
        ]
        if wiktionary:
            args.append("--wiktionary")

        AmiLib().run_command(args)
        assert Path(output_dict).exists()

    def test_make_dict_with_figures(self):
        stem = "with_figures1"
        output_dict = Path(Resources.TEMP_DIR, "words", "html", f"{stem}.html")

        words = [
            "albedo",
            "ablation",
        ]
        args = [
            "DICT",
            "--words", words,
            "--dict", output_dict,
            "--title", stem,
            "--figures", "wikipedia",
        ]
        AmiLib().run_command(args)
        assert Path(output_dict).exists()

    def test_disambiguation(self):
        stem = ("disambiguation")
        output_dict = Path(Resources.TEMP_DIR, "words", "html", f"{stem}.html")

        args = [
            "DICT",
            "--words",
            "stubble",
            "anthropogenic",
            "--dict", output_dict,
            # "--figures",
        ]
        AmiLib().run_command(args)
        assert Path(output_dict).exists()

    def test_create_dict_with_wikimedia_descriptions(self):
        """
        from commandline add descriptions to ditionary entries from wikimedia sources
        """
        stem = "parijat"
        output_dict = Path(Resources.TEMP_DIR, "words", "html", f"{stem}.xml")

        args = [
            "DICT",
            "--words",
            stem,
            "--description",
            "wikipedia",
            "wikidata",
            "wiktionary",
            "--dict", output_dict,
        ]
        AmiLib().run_command(args)
        assert Path(output_dict).exists()


class IPCCDictTest(AmiAnyTest):
    def test_ipcc_dictionaries_from_URL(self):
        """
        read a number of IPCC dictionaries from URL and test their validity.
        Maybe later move code to dictionary project
        """
        WG3_ROOT = "https://raw.githubusercontent.com/petermr/semanticClimate/main/ipcc/ar6/wg3"
        dictionary_names = [
            "Chapter02/dict/ip_3_2_emissions_abb.xml", "Chapter07/dict/ip_3_7_agri_abb.xml",
            "wg2_03/dict/ip_3_3_longmitig_abb.xml", "Chapter08/dict/ip_3_8_urban_abb.xml",
            "Chapter05/dict/ip_3_5_socmitig_abb.xml", "Chapter17/dict/ip_3_17_sustdev_abb.xml",
            "Chapter06/dict/ip_3_6_energy_abb.xml"
        ]

        for dictionary_name in dictionary_names:
            url = WG3_ROOT + "/" + dictionary_name
            print(f" url {url}")
            try:
                dictionary = AmiDictionary.create_dictionary_from_url(url)
            except Exception as e:
                print(f"cannot download dictionary {url} beacuse {e}")

    @unittest.skipUnless(Path(IPCC_DICT_ROOT).exists(), f"requires checked out IPCC dictionaries {IPCC_DICT_ROOT}")
    def test_ipcc_dictionaries_from_file(self):
        """
        read a number of IPCC dictionaries from LOCAL (PMR-only) files and test their validity.
        Maybe later move codeto dictionary project
        """
        dictionary_names = [
            "Chapter02/dict/ip_3_2_emissions_abb.xml",
            "Chapter07/dict/ip_3_7_agri_abb.xml",
            # "wg2_03/dict/ip_3_3_longmitig_abb.xml",
            "Chapter08/dict/ip_3_8_urban_abb.xml",
            "Chapter05/dict/ip_3_5_socmitig_abb.xml",
            "Chapter17/dict/ip_3_17_sustdev_abb.xml",
            "Chapter06/dict/ip_3_6_energy_abb.xml"
        ]
        for dictionary_name in dictionary_names:
            dict_file = Path(IPCC_DICT_ROOT, dictionary_name)
            print(f"reading {dict_file}")
            if not dict_file.exists():
                logger.warning(f"dict file does not exist {dict_file}")
                continue
            try:
                dictionary = AmiDictionary.read_dictionary(dict_file)
                assert dictionary is not None, f"reading from {dict_file}"
                assert dictionary.get_entry_count() > 0
            except Exception as e:
                print(f"cannot read dictionary {dict_file} beacuse {e}")

# ======================= helpers =====================
    def teardown(self):
        dict1_root = None



def main(argv=None):
    print(f"running PDFArgs main")
    pdf_args = AmiDictArgs()
    try:
        pdf_args.parse_and_process()
    except Exception as e:
        print(traceback.format_exc())
        print(f"***Cannot run pyami***; see output for errors: {e}")


if __name__ == "__main__":
    main()
else:
    pass


def main():
    pass



import argparse
import json
from abc import ABC, abstractmethod
from collections import Counter
import logging
from pathlib import Path

from amilib.util import Util

"""Base class for specialist sub_parsers .
Also includes ArgParseBuilder
"""

logger = Util.get_logger(__name__)
logger.setLevel(logging.INFO)

OPERATION = "operation"
class AbstractArgs(ABC):

    PIPELINE = "Pipeline"
    DEBUG = "debug"
    INDIR = "indir"
    INPUT = "input"
    INFORMAT = "informat"
    KWARGS = "kwords"
    OPERATION = "operation"
    OUTPUT = "output"
    OUTDIR = "outdir"

    DEBUG_HELP = f"will output debugging information (not fully implemented) \n"

    INPUT_HELP = f"input from:\n" \
                 f"   file/s single, multiple, and glob/wildcard (experimental)\n" \
                 f"   directories NYI\n" \
                 f"   URL/s (must start with 'https:'); provide {OUTDIR} for output \n"

    INDIR_HELP = f"Directory containing input files\n"

    OPERATION_HELP = "operation to be performed (--operation is normally required)"
    OUTPUT_HELP = "output file or similar resource"

    OUTDIR_HELP = "output directory, required for URL input. If not given, autogenerated from file names"

    def get_input_help(self):
        return self.INPUT_HELP

    def get_indir_help(self):
        return self.INDIR_HELP

    def get_output_help(self):
        return self.OUTPUT_HELP

    def get_outdir_help(self):
        return self.OUTDIR_HELP

    def __init__(self):
        self.operation = None
        self.parser = None
        self.parsed_args = None
        self.ref_counter = Counter()
        self.arg_dict = self.create_default_arg_dict()
        self.subparser_arg = "UNKNOWN"

    def create_arg_dict(self, args=None):
        """
        takes args or self.parsed_args as a list of items(tuples) and creates a dictionary (self.arg_dict)
        :param args: list of key-values
        :return: dict() created form this
        """
        logger.debug(f"create_arg_dict args: {args}")
        if args:
            self.parsed_args = args
        if not self.parsed_args:
            return None
        try:
            arg_vars = vars(self.parsed_args)
        except TypeError:
            arg_vars = self.parsed_args
        self.arg_dict = dict()
        for item in arg_vars.items():
            key = item[0]
            if item[1] is None:
                pass
            elif type(item[1]) is list and len(item[1]) == 1:
                self.arg_dict[key] = item[1][0]
            else:
                self.arg_dict[key] = item[1]

        return self.arg_dict

    def _parse_and_process1(self, argv_):
        """
        Takes args from CLI , creates self.arg_dict, and passes to self.process_args()
        :param argv_: name-value pairs usually from CLI

        """
        # logger.debug("running parse_and_process1 in util?")
        # logging.warning(f"********** args for parse_and_process1 {argv_}")
        self.parsed_args = argv_ if self.parser is None else self.parser.parse_args(argv_)
        self.arg_dict = self.create_arg_dict()
        self.process_args()

    def add_arguments(self):
        """
        Not used
        maybe this should be called by subclasses
        """

        self.parser.add_argument(f"--{self.DEBUG}",
                                 action='store_true',
                                 help=self.DEBUG_HELP)

        self.parser.add_argument(f"--{self.INPUT}", nargs="+",
                                 help=self.INPUT_HELP)

        self.parser.add_argument(f"--{self.INDIR}", nargs="+",
                                 help=self.INDIR_HELP)

        self.parser.add_argument(f"--{self.OUTPUT}", nargs="+",
                                 help=self.OUTPUT_HELP)

        self.parser.add_argument(f"--{self.OUTDIR}", nargs="+",
                                 help=self.OUTDIR_HELP)

        self.parser.add_argument(f"--{self.OPERATION}", nargs="+",
                                 help=self.OPERATION_HELP)

        INFORM_HELP = "input format/s; experimental"
        self.parser.add_argument(f"--{self.KWARGS}", nargs="*",
                help="space-separated list of colon_separated keyword-value pairs, "
                     "format kw1:val1 kw2:val2;\nif empty list gives help")


    @abstractmethod
    def process_args(self):
        pass

    @abstractmethod
    def create_default_arg_dict(self):
        pass

    @abstractmethod
    def add_arguments(self):
        pass

    # @property
    # def module_stem(self):
    #     """name of module"""
    #     return Path(__file__).stem
    #
    def get_operation(self):
        """
        NOT USED, maybe it should be
        The operation to run (makes this explicit)
        """
        operation = self.arg_dict.get(AbstractArgs.OPERATION)
        return operation

    def get_indir(self):
        indir = self.arg_dict.get(AbstractArgs.INDIR)
        return indir

    # are these used?
    def get_input(self):
        inputx = self.arg_dict.get(AbstractArgs.INPUT)
        return inputx

    def get_outdir(self):
        outdir = self.arg_dict.get(AbstractArgs.OUTDIR)
        return outdir

    def get_output(self):
        output = self.arg_dict.get(AbstractArgs.OUTPUT)
        return output


    """
    These kwargs methods may by used by subclasses
    """
    def parse_kwargs_to_string(self, kwargs, keys=None):

        kwargs_dict = {}
        logger.info(f"args: {kwargs}")
        if not kwargs:
            if keys:
                logger.debug(f"possible keys: {keys}")
        else:
            if type(kwargs) is not list:
                kwargs = [kwargs]
            for arg in kwargs:
                logger.debug(f"pair {arg}")
                argz = arg.split(':')
                key = argz[0].strip()
                value = argz[1].strip()
                kwargs_dict[key] = value
            logger.warning(f"kwargs_dict {kwargs_dict}")
        return kwargs_dict


    def get_kwargs(self):
        """
        used if we ever start using **kwargs,
        currently not used
        """
        kwargs = self.arg_dict.get(AbstractArgs.KWARGS)
        logger.debug(f"kwargs {kwargs}")
        if kwargs is None:
            return None
        if len(kwargs) == 0:
            self.kwargs_help()
        else:
            pass

        return


    def kwargs_help(self):
        """
        only used by **kwargs options
        """
        logger.debug(f"key value pairs separated by ':' ; normally explicitly offered by subclass ")


    @classmethod
    def make_sub_parser(cls, subclass, subparsers):
        """make subparser from subparsers
        requires self.subparser_arg (probably should be argument
        ALSO adds arguments through `self.add_arguments
        :param subclass: name of the subclass`
        :param subparsers: list of subparsers (to append subparser to)
        :return: new subparser"""
        subclass.parser = subparsers.add_parser(subclass.subparser_arg)
        logger.debug(f"subclass_parser for {subclass} is {subclass.parser}")
        subclass.add_arguments()
        return subclass.parser

class ArgParseBuilder:
    """
    Not sure whether this is necessary but there is a test using it
    """
    ARG_LIST = "arg_list"
    DESCRIPTION = "description"

    def __init__(self):
        self.parser = None

    def create_arg_parse(self, arg_dict=None, arg_dict_file=None):
        # arg_dict_file takes precedence
        if arg_dict_file and arg_dict_file.exists():
            with open(arg_dict_file, 'r') as f:
                data = f.read()
                arg_dict = json.loads(data)
                logger.debug(f"arg_dict {arg_dict}")

        if arg_dict is not None:
            desc = f'{arg_dict.get(self.DESCRIPTION)}'
            logger.debug(f"\ndesc: '{desc}'")
            self.parser = argparse.ArgumentParser(description=desc)
            arg_list = arg_dict.get(self.ARG_LIST)
            if arg_list is None:
                raise ValueError(f"must give arg_list to ArgParseBuilder")
            for arg_dict in arg_list:
                if not type(arg_dict) is dict:
                    raise ValueError(f"arg_list_dict {arg_dict} is not a dict")
                args = arg_dict.keys()
                for arg in args:
                    logger.debug(f"\n{arg}:")
                    param_dict = arg_dict.get(arg)
                    self.process_params(param_dict)
                # self.parser.add_argument(f"--{ProjectArgs.PROJECT}", type=str, nargs=1, help="project directory")

    """https://stackoverflow.com/questions/28348117/using-argparse-and-json-together"""

    def process_params(self, param_dict):
        for param, param_val in param_dict.items():
            logger.debug(f"  {param}='{param_val}'")


class AmiArgParseException(Exception):
    """
    to capture error messages from AmiArgparser
    """
    pass


class AmiArgParser(argparse.ArgumentParser):
    """
    subclasses ArgumentParser and overrides error()
    """

    def error(self, message):
        """
        raises self.exit(2, error_message) so can be caught
        """
        raise AmiArgParseException(message)





import argparse
import logging
import textwrap
from pathlib import Path

from amilib.ami_args import AbstractArgs, AmiArgParser
from amilib.ami_dict import AmiDictionary
from amilib.util import Util
from amilib.ami_html import (HtmlLib)

ANNOTATE = "annotate"
DICT = "dict"
INDEX = "index"
INPATH = "inpath"
NOINPUTSTYLES = "no_input_styles"
OPERATION = "operation"
OUTPATH = "outpath"
SEARCH = "SEARCH"
TITLE = "title"

UNKNOWN = "unknown"

logger = Util.get_logger(__name__)
logger.setLevel(logging.INFO)

class SearchArgs(AbstractArgs):
    pass

    def __init__(self):
        super().__init__()
        self.dictfile = None
        self.operation = None
        self.parser = None
        self.title = UNKNOWN
        self.words = None

        self.subparser_arg = SEARCH


    def add_arguments(self):
        """creates adds the arguments for pyami commandline

        """
        logger.debug(f"================== add arguments SEARCH ================")
        if self.parser is None:
            self.parser = AmiArgParser(
                usage="SEARCH: amilib always uses subcommands (HTML,PDF, DICT, SEARCH)\n e.g. amilib SEARCH --help"
            )

        self.parser.description = textwrap.dedent(
            'SEARCH tools. \n'
            '----------\n'
            'Search documents and corpora and make indexes and maybe knowledge graphs.'
            'Not yet finished.\n'
            '\nExamples:\n'
            '  * SEARCH --help\n'
        )
        self.parser.formatter_class = argparse.RawDescriptionHelpFormatter

        self.parser.add_argument("--debug", type=str,
                                 help="debug these during parsing (NYI)")
        self.parser.add_argument(f"--{DICT}", type=str, nargs=1,
                                 help="path for dictionary  *.xml or *.html)")
        self.parser.add_argument(f"--{INPATH}", type=str, nargs="+", help="path for input file(s)")
        self.parser.add_argument(f"--{OPERATION}", type=str, nargs="+",
                                 default=ANNOTATE,
                                 choices=[ANNOTATE, INDEX, NOINPUTSTYLES],
                                 help=f"operation: "
                                      f"'{NOINPUTSTYLES}' needs '{INPATH} ; remove styles from inpath\n"
                                      f"'{ANNOTATE}' needs '{INPATH} and {DICT}'; annotates words/phrases\n"
                                      f"'{INDEX}' needs '{INPATH}' optionally {OUTPATH} (NYI)\n"
                                      f" default = {ANNOTATE}"
                                 )
        self.parser.add_argument(f"--{OUTPATH}", type=str, nargs="+",
                                 help="output file ")
        self.parser.add_argument(f"--{TITLE}", type=str,
                                 default="unknown",
                                 help="internal title for dictionary, normally same as stem of dictionary file")
        self.parser.epilog = """
        Examples:
        SEARCH --operation annotate --inpath infile --dict dictfile --outpath outfile  # uses dictfile to annotate words in infile and write outfile\n
        """
        self.parser.epilog = "============ SEARCH epilog ==========="
        return self.parser

    # class PDFArgs:
    def process_args(self):
        """runs parsed args
        pass
        """
        logger.debug(f"SEARCH process_args {self.arg_dict}")
        if not self.arg_dict:
            logger.debug(f"no arg_dict given, no actiom")

        self.dictfile = self.arg_dict.get(DICT)
        self.inpath = self.arg_dict.get(INPATH)
        self.outpath = self.arg_dict.get(OUTPATH)
        self.operation = self.arg_dict.get(OPERATION)
        self.title = self.arg_dict.get(TITLE)
        logger.info(f"read arguments\n"
                    f"inpath: {self.inpath}\n"
                    f"dictfile: {self.dictfile}\n"
                    f"outpath: {self.outpath}\n"
                    f"operation: {self.operation}\n"
                    f"title: {self.title}\n"
                    )
        if self.operation is None:
            logger.warning("No operation given")
            return
        self.remove_input_styles = NOINPUTSTYLES in self.operation

        if ANNOTATE in self.operation:
            self.markup_file_with_dict()

        if self.operation == INDEX:
            self.make_index()

# output goes here
        pass



    @classmethod
    def create_default_arg_dict(cls):
        """returns a new COPY of the default dictionary
        """
        arg_dict = dict()
        return arg_dict

    def markup_file_with_dict(self):
        """
        uses dictionary from self.dictfile to markup self.inpath and write to self.outpath:
        """
        logger.debug(f"search {self.inpath} with {self.dictfile} and output to {self.outpath}")
        if self.inpath and self.dictfile and self.outpath:
            self.make_dictionary_markup_file(self.inpath, self.dictfile, self.outpath)

    def make_dictionary_markup_file(self, inpath, dictfile, outpath):
        """
        create dictionary and annotate file
        :param inpath: HTML file to annotate
        :param dictfile: file containing dictionary
        :param outpath: output for annotated file
        TODO allow for more dictionaries
        """
        from amilib.ami_dict import AmiDictionary

        dictfile = str(dictfile)

        if not dictfile.endswith(".html"):
            logger.error(f"dictionary for commandline must be HTML")
            return None
        # TODO this should not be in AmiDictionary
        AmiDictionary.read_html_dictionary_and_markup_html_file(
            str(inpath), str(outpath), remove_styles=self.remove_input_styles,  html_dict_path=dictfile)
        # logger.info(f"wrote annotated file {outpath}")

    @classmethod
    def read_html_dictionary_and_markup_html_file(cls, inpath, outpath, html_dict_path):
        """
        read semantic HTML file, extract paras with ids, create AmiDictionary from HTML,
        markup paras, and write marked  file
        :param inpath: to be marked up
        :param outpath: resulting marked file
        :param html_dict_path: dictiomary in HTML format
        :return: HTML element marked_up
        """
        assert Path(inpath).exists()
        paras = HtmlLib._extract_paras_with_ids(inpath)
        assert Path(html_dict_path).exists()
        # TODO this should not be an AmiDictinary responsibility. Use the phrases
        dictionary = AmiDictionary.create_from_html_file(html_dict_path)
        assert dictionary is not None
        phrases = dictionary.get_terms()
        dictionary.location = html_dict_path
        HtmlLib.search_phrases_in_paragraphs(paras, phrases, markup=html_dict_path)
        # write marked_up html
        chapter_elem = paras[0].xpath("/html")[0]
        HtmlLib.write_html_file(chapter_elem, outpath, debug=True)
        assert Path(outpath).exists()
        return chapter_elem






import argparse
import json
import logging
import traceback
from collections import Counter
from pathlib import Path

# cyclic import
from amilib.ami_args import AbstractArgs
# from amilib.ami_dict import AmiDictionary # CYCLIC import
# from amilib.ami_dict import AmiEntry
from amilib.file_lib import FileLib
from amilib.util import Util
from amilib.wikimedia import WikidataPage, WikidataLookup, WiktionaryPage
from amilib.ami_html import HtmlLib

# commandline
DESCRIPTION = "description"
DICT = "dict"
FIGURES = "figures"
IMAGE = "image"
INPATH = "inpath"
OPERATION = "operation"
OUTPATH = "outpath"
SYNONYM = "synonym"
TITLE = "title"
VALIDATE = "validate"
WORDS = "words"

UNKNOWN = "unknown"

# operations
CREATE_DICT = "create"
EDIT_DICT = "edit"
MARKUP_FILE = "markup"
VALIDATE_DICT = "validate"

# description sources
WIKIPEDIA = "wikipedia"
WIKIDATA = "wikidata"
WIKTIONARY = "wiktionary"

logger = Util.get_logger(__name__)
logger.setLevel(logging.DEBUG)


class AmiDictArgs(AbstractArgs):
    """Parse args to build and edit dictionary"""

    def __init__(self):
        """arg_dict is set to default"""
        logger.debug("creating HTML Args")
        super().__init__()
        self.dictfile = None
        self.indir = None
        self.inpath = None
        self.outpath = None
        self.outstem = None
        self.outdir = None
        self.arg_dict = None
        self.subparser_arg = "DICT"
        self.ami_dict = None

    def add_arguments(self):
        """
        add arguments into self.parser
        """

        if self.parser is None:
            self.parser = argparse.ArgumentParser()
        """adds arguments to a parser or subparser"""
        self.parser.description = 'AMI dictionary creation, validation, editing'
        self.parser.add_argument(f"--{DESCRIPTION}", type=str, nargs="+",
                                 choices=[WIKIPEDIA, WIKTIONARY, WIKIDATA],
                                 help="add extended description tp dict from one or more of these")
        self.parser.add_argument(f"--{DICT}", type=str, nargs=1,
                                 help="path for dictionary (existing = edit; new = create (type depends on suffix *.xml or *.html)")
        self.parser.add_argument(f"--{INPATH}", type=str, nargs="+", help="path for input file(s)")
        self.parser.add_argument(f"--{FIGURES}", type=str,
                                 nargs="*",
                                 default="None",
                                 choices=["None", WIKIPEDIA, WIKIDATA],
                                 help=f"sources for figures: "
                                      f"'{WIKIPEDIA}' uses infobox or first thumbnail, {WIKIDATA} uses first figure"
                                 )
        self.parser.add_argument(f"--{OPERATION}", type=str,
                                 default=CREATE_DICT,
                                 choices=[CREATE_DICT, EDIT_DICT, MARKUP_FILE, VALIDATE_DICT],
                                 help=f"operation: "
                                      f"'{CREATE_DICT}' needs '{WORDS}'\n"
                                      f" '{EDIT_DICT}' needs '{INPATH}'\n"
                                      f" '{MARKUP_FILE}' need '{INPATH}' and '{OUTPATH}` (move to search?)\n"
                                      f" '{VALIDATE}' requires '{INPATH}'\n"
                                      f" default = '{CREATE_DICT}"
                                 )
        self.parser.add_argument(f"--{OUTPATH}", type=str, nargs="+",
                                 help="output file ")
        self.parser.add_argument(f"--{SYNONYM}", type=str, nargs="+",
                                 help="add sysnonyms (from Wikidata) for terms (NYI)")
        self.parser.add_argument(f"--{TITLE}", type=str,
                                 default="unknown",
                                 help="internal title for dictionary, normally same as stem of dictionary file")
        self.parser.add_argument(f"--{VALIDATE}", action="store_true", help="validate dictionary; DEPRECATED use '--operation validate'")
        self.parser.add_argument(f"--{WIKIDATA}", type=str, nargs="*", help=f"DEPRECATED use --description {WIKIDATA} add WikidataIDs (NYI)")
        self.parser.add_argument(f"--{WIKIPEDIA}", type=str, nargs="*",
                                 help="add Wikipedia link/s; DEPRECATED use '--description wikipedia'")
        self.parser.add_argument(f"--{WIKTIONARY}", type=str, nargs="*",
                                 help="add Wiktionary output as html (may be messy); DEPRECATED use '--description wiktionary'")
        self.parser.add_argument(f"--{WORDS}", type=str, nargs="*",
                                 help="path/file with words or list of words to create dictionaray")
        self.parser.epilog = """
        Examples:
        DICT --words wordsfile --dict dictfile --description wikipedia   # creates dictionary from wordsfile and adds wikipedia info\n
        
        
        """

    # class AmiDictArgs:
    def process_args(self):
        """runs parsed args
        :return:
        """
        logger.debug(f"DICT process_args {self.arg_dict}")
        if not self.arg_dict:
            logger.debug(f"no arg_dict given, no actiom")

        self.description = self.arg_dict.get(DESCRIPTION)
        self.dictfile = self.arg_dict.get(DICT)
        self.figures = self.arg_dict.get(FIGURES)
        if self.figures == "None":
            self.figures = None
        self.inpath = self.arg_dict.get(INPATH)
        self.outpath = self.arg_dict.get(OUTPATH)
        self.operation = self.arg_dict.get(OPERATION)
        self.synonym = self.arg_dict.get(SYNONYM)
        self.title = self.arg_dict.get(TITLE)
        self.validate = self.arg_dict.get(VALIDATE)
        self.wikidata = self.arg_dict.get(WIKIDATA)
        self.wikipedia = self.arg_dict.get(WIKIPEDIA)
        self.wiktionary = self.arg_dict.get(WIKTIONARY)
        self.words = self.make_input_words( self.arg_dict.get(WORDS))


        if self.operation is None:
            raise ValueError("No operation given")
        elif self.operation == CREATE_DICT:
            self.create_dictionary_from_words(self.title)
        elif self.operation == EDIT_DICT:
            self.edit_dictionary()
        elif self.operation == MARKUP_FILE:
            logger.warning(f"DEPRECATED; use 'amilib SEARCH'")
            self.markup_file_with_dict()
        elif self.operation == VALIDATE:
            self.validate_dict()
        else:
            raise ValueError(f"unknown oeration {self.operation}")

        if self.dictfile:
            if self.ami_dict:
                self.ami_dict.create_html_write_to_file(self.dictfile, debug=True)

    def add_descriptions_old(self):
        if self.wikidata is not None:
            hit_dict = self.add_wikidata_to_dict()
            status = self.validate_dict()
        # for argument --wikipedia
        if self.wikipedia is not None:
            logger.debug(f"Wikipedia lookup")
            hit_dict = self.add_wikipedia_to_dict()
            status = self.validate_dict()
        # for argument --wikipedia
        if self.wiktionary is not None:
            logger.debug(f"Wiktionary lookup {self.wiktionary}")
            if len(self.wiktionary) > 0:
                logger.debug(f"searching wiktionary for {self.wiktionary}")
                htmlx = WiktionaryPage.search_terms_create_html(self.wiktionary)
                temp = Path(Path(__file__).parent.parent, "temp")
                logger.info(f"temp dir {temp}")
                HtmlLib.write_html_file(htmlx, Path(temp, "wiktionary", f"{self.title}.html"))
                body = htmlx
            else:
                logger.debug(f"add to dictionary NYI")

    def add_descriptions(self):
        """
        add 1 or more descriptons from WIKIPEDIA, WIKIDATA or WIKTIONARY
        """
        if self.description is None:
            logger.warning("No entry description given")
            return
        for entry_description in self.description:
            if entry_description == WIKIDATA:
                hit_dict = self.add_wikidata_to_dict()
                status = self.validate_dict()
            elif entry_description == WIKIPEDIA:
                logger.debug(f"Wikipedia lookup")
                hit_dict = self.add_wikipedia_to_dict()
                status = self.validate_dict()
            elif entry_description == WIKTIONARY:
                logger.debug(f"Wiktionary lookup {self.wiktionary}")
                logger.debug(f"searching wiktionary for {self.wiktionary}")
                htmlx = WiktionaryPage.search_terms_create_html(self.wiktionary)
                temp = Path(Path(__file__).parent.parent, "temp")
                logger.info(f"temp dir {temp}")
                HtmlLib.write_html_file(htmlx, Path(temp, "wiktionary", f"{self.title}.html"))
                body = htmlx

    def create_dictionary_from_words(self, title):
        """
        use self.words to create a dictionary
        potentially add decriptions and figures
        :param title: mandatory title
        """
        if title is None:
            logger.error("must give title")
            return
        if self.words is None:
            logger.warning("no words given in args")
            return

        self.build_or_edit_dictionary()
        assert self.ami_dict is not None

        self.add_descriptions()
        if self.figures is not None:
            self.add_figures()

        if self.dictfile is not None:
            self.ami_dict.create_html_write_to_file(self.dictfile, debug=True)

    # class AmiDictArgs:

    @classmethod
    def create_default_arg_dict(cls):
        """returns a new COPY of the default dictionary"""
        """
        is this used?
        """
        # from amilib.dict_args import DICT, VALIDATE, WIKIDATA, WORDS
        arg_dict = dict()
        arg_dict[DICT] = None
        arg_dict[VALIDATE] = None
        arg_dict[WIKIDATA] = None
        arg_dict[WORDS] = None
        return arg_dict

    def build_or_edit_dictionary(self):
        # cyclic imports TODO
        from amilib.ami_dict import AmiDictionary

        if self.words is not None:
            logger.info(f"creating dictionary from {self.words[:3]}...")
            self.ami_dict, _ = AmiDictionary.create_dictionary_from_words(
                terms=self.words, title=self.title, wiktionary=False)
        return self.ami_dict

    def add_wikidata_to_dict(self, description_regex=None):

        # TODO fix circular imports

        desc_counter = Counter()
        hit_dict = dict()
        if self.dictfile is None or self.ami_dict is None:
            logger.debug(f"add_wikidata_to_dict requires existing dictionary")
            return
        wikidata_lookup = WikidataLookup()
        for entry in self.ami_dict.entries:
            self.add_wikidata_to_entry(entry, hit_dict, wikidata_lookup)

        logger.debug(f"JSON DUMPS {json.dumps(hit_dict, sort_keys=False, indent=2)}")
        counter = Counter()

        """
{
  "acetone": {
    "Q49546": {
      "title": "acetone",
      "description": "chemical compound",
      "score": 0
    },
    "Q222936": {
      "title": "acetone cyanohydrin",
      "description": "chemical compound",
      "score": 1
    },
  },
  "benzene": {
    ...
    },
  },
}
        """
        for hit_dict_key in hit_dict.keys():
            subdict = hit_dict[hit_dict_key]
            logger.debug(f"sub-title {subdict.keys()}")
            for key, item in subdict.items():
                logger.debug(f".....{key} item_title {item['title']}")

        return hit_dict

    def add_wikidata_to_entry(self, entry, hit_dict, wikidata_lookup):
        # cyclic import
        from amilib.ami_dict import AmiEntry
        ami_entry = AmiEntry.create_from_element(entry)
        term = entry.attrib["term"]
        term_dict = dict()
        hit_dict[term] = term_dict
        qitem0, desc, qitem_hits = wikidata_lookup.lookup_wikidata(term)
        if qitem_hits is None:
            logger.error(f"no qitem_hits {term} in add_wikidata_to_dict")
            return
        for i, qid in enumerate(qitem_hits):
            qitem_hit_dict = self.create_hit_dict_for(i, qid)
            term_dict[qid] = qitem_hit_dict
            ami_entry.add_hits_to_xml(qid, qitem_hit_dict)

    def add_wikipedia_to_dict(self):
        # TODO CYCLIC import
        from amilib.ami_dict import AmiEntry
        """
        NYI
        """
        if self.ami_dict.entries is None:
            logger.error("No self.ami_dict.entries")
            return
        for entry_elem in self.ami_dict.entries:
            ami_entry = AmiEntry.create_from_element(entry_elem)
            ami_entry.lookup_and_add_wikipedia_page()

    def create_hit_dict_for(self, serial, qitem_hit):
        qitem_hit_dict = dict()
        page = WikidataPage(pqitem=qitem_hit)
        description = page.get_description()
        title = page.get_title()
        qitem_hit_dict["title"] = title
        qitem_hit_dict["description"] = description
        qitem_hit_dict["score"] = serial

        return qitem_hit_dict

    @property
    def module_stem(self):
        """name of module"""
        return Path(__file__).stem

    def make_input_words(self, words):
        self.words = None if words is None else FileLib.get_input_strings(words)
        return self.words

# ========== create=======

    def add_figures(self):
        # TODO CYCLIC import
        from amilib.ami_dict import AmiEntry
        """
        iterate over all entries, and find the WikipediaPage. 
        extract figure/s from the page and add to the entry 
        calls ami_entry.add_figures_to_entry
        """
        if self.ami_dict.entries is None:
            logger.warning("No self.ami_dict.entries")
            return
        for entry_elem in self.ami_dict.entries:
            ami_entry = AmiEntry.create_from_element(entry_elem)
            wikipedia_page = ami_entry.lookup_and_add_wikipedia_page()
            # ami_entry.add_figures_to_entry_old()
            if wikipedia_page is not None:
                ami_entry.add_figures_to_entry(wikipedia_page)

    # ========== eiit ========
    def edit_dictionary(self):
        self.add_descriptions()


# ========== markup ========
    def markup_file_with_dict(self):
        """
        uses dictionary from self.dictfile to markup self.inpath and write to self.outpath:
        """
        logger.warning("DEPRECATED markup_file will be moved to SEARCH")
        if self.inpath and self.dictfile and self.outpath:
            self.make_dictionary_markup_file(self.inpath, self.dictfile, self.outpath)

    def make_dictionary_markup_file(self, inpath, dictfile, outpath):
        """
        create dictionary and annotate file
        :param inpath: HTML file to annotate
        :param dictfile: file containing dictionary
        :param outpath: output for annotated file
        TODO allow for more dictionaries
        """
        logger.warning("DEPRECATED: use 'amilib SEARCH'")
        from amilib.ami_dict import AmiDictionary

        dictfile = str(dictfile)

        if not dictfile.endswith(".html"):
            logger.error(f"dictionary for commandline must be HTML")
            return None
        # dictionary = AmiDictionary.create_from_html_file(dictfile)
        # # logger.info(f"terms: {len(dictionary.get_terms())} {dictionary.get_terms()[0]}")
        # dictionary.markup_html_from_dictionary(inpath, outpath)
        AmiDictionary.read_html_dictionary_and_markup_html_file(
            str(inpath), str(outpath), html_dict_path=dictfile)

    # ========== validate ========
    def validate_dict(self):
        logger.debug(f"VALIDATING {self.ami_dict}")
        status = False
        if self.dictfile and Path(self.dictfile).exists():
            self.ami_dict.check_validity()
        else:
            logger.error(f"vallidate_dict requires existing dictionary; no validation")
        logger.debug(f"validation finished")
        return status



# =============================


# ====================


def main(argv=None):
    # AMIDict.debug_tdd()
    logger.info(f"running AmiDict main")
    dict_args = AmiDictArgs()
    try:
        dict_args.parse_and_process()
    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.error(f"***Cannot run amidict***; see output for errors: {e} ")


if __name__ == "__main__":
    logger.info("running dict main")
    main()
else:

    #    logger.debug("running dict main anyway")
    #    main()
    pass



"""tests for file_lib"""
import glob
import os
import logging
import unittest

from pathlib import Path

from amilib.file_lib import FileLib
from amilib.util import Util

from test.resources import Resources
from test.test_all import AmiAnyTest

FILE_LIB = "file_lib"
PDF_LIB = "pdf_lib"
TEXT_LIB = "text_lib"
_SETUP = "_setup"
_TEARDOWN = "_teardown"
TEST = "test"

"""
a few tests on globbing, etc"""

logger = Util.get_logger(__name__)

class File0Test(AmiAnyTest):
    # nothing much here
    logger = logging.getLogger("test_file")
    TEST = "test"

    OPTIONS = [
        FILE_LIB,
        PDF_LIB,
        TEXT_LIB,
        _SETUP,
        _TEARDOWN,
    ]

    @classmethod
    def example_setup(cls, pyamix):
        """ setup test or examples
        these are general commands (i.e. not subcommands)
        :pyamix:
        """
        pyamix.run_commands([
            "--delete ${exam_temp}",
            "--copy ${examples_test.p} ${exam_temp} overwrite",
        ])

    @classmethod
    def example_teardown(cls, pyamix):
        """ clean example files

        """
        pyamix.run_command([
            "--delete", "${exam_temp}",
        ])

    @classmethod
    def run_arg_tests(cls, args):
        """This needs revision , maybe using Examples()"""
        cls.logger.warning(f"*****running tests : {args[TEST]}")
        if not args[TEST]:
            cls.logger.warning(f"No tests given: choose some/all of {TEST}")
            return
        if FILE_LIB in args[TEST]:
            cls.logger.warning("run test_file")
            cls.test_file.main()
        if PDF_LIB in args[TEST]:
            cls.logger.warning("run test_pdf")
            cls.test_pdf.test_read_pdf()
        if TEXT_LIB in args[TEST]:
            cls.logger.warning("run test_text NYI")


    @unittest.skip("if not windows")
    def test_glob_with_windows_FILE(self):
        """
        some instances of glob.glob with fully forward slashes generate backslash or double backslash on windows
        """
        input_dir = Path(Resources.TEST_RESOURCES_DIR, "ar6", "cleaned_content")
        assert input_dir.exists()
        glob_str = f"{str(input_dir)}/**/html_with_ids.html"
        files = FileLib.posix_glob(glob_str, recursive=True)
        assert len(files) > 70
        for file in files[:5]:
            logger.debug(f"file: {file}")
            assert not "\\" in file # backslash is not generated by UNIX/MacOS
            assert "/" in file


    @classmethod
    @unittest.skip("not yet right")
    def test_expand_braces(cls):
        """
        no longer tests braces
        :return:
        """

        home = os.path.expanduser("~")
        file = __file__
        python_dir = os.path.abspath(file + "/../")
        resources_dir = os.path.join(python_dir, "resources")
        logger.debug("python", os.path.abspath(python_dir))
        physchem_dir = os.path.abspath(python_dir + "/../")
        open_diagram = os.path.abspath(physchem_dir + "/../")
        logger.debug(os.path.abspath(open_diagram))
        opend = os.path.join(open_diagram, "*")
        logger.debug("opend", opend)
        logger.debug(f"od0 {glob.glob(opend)}")
        open_diagram01 = os.path.join(physchem_dir, "**", "liion10", "**", "*.*ml")
        logger.debug(f"od01 {glob.glob(open_diagram01)}")

        # open_diagram02 = os.path.join(home, "projects", "*iagram", "{*ot*,*.md}")
        # logger.debug("od02", bg.braced_glob(open_diagram02))
        open_diagram11 = os.path.join(home, "projects", "*iagram", "*", "*", "*.xml")
        logger.debug("od", open_diagram11)
        logger.debug("b", glob.glob(open_diagram11))
        open_diagram12 = os.path.join(home, "projects", "*iagram", "**", "*.xml")
        logger.debug("od", open_diagram12)

    def test_get_input_strings(self):
        """
        reads lists of words with optional split
        """
        words = FileLib.get_input_strings("foo")
        assert words == ["foo"]
        words = FileLib.get_input_strings(["foo"])
        assert words == ["foo"]
        words = FileLib.get_input_strings(["foo", "bar"])
        assert words == ["foo", "bar"]
        words = FileLib.get_input_strings("foo bar")
        assert words == ["foo bar"]
        words = FileLib.get_input_strings("foo bar", split=True)
        assert words == ["foo", "bar"]

    def test_get_input_strings_from_file(self):
        """
        reads lists of words in file/s
        """
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_2.txt"))
        assert words == ["anthropogenic", "tropospheric"]
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt"))
        assert words == [
            'anthropogenic', 'physical-biogeochemical', 'peat drainage', 'tropospheric', 'permafrost',
            'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating'
        ]
        words = FileLib.get_input_strings(Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt"), split=True)
        # assert words == [
        #     'anthropogenic', 'physical-biogeochemical', 'peat', 'drainage', 'tropospheric', 'permafrost',
        #     'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating'
        # ]
        assert words ==['anthropogenic',
 'physical-biogeochemical',
 'peat',
 'drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']
        # list of files
        words = FileLib.get_input_strings([
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_5.txt"),
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt")])

        # assert words == ['anthropogenic', 'physical-biogeochemical', 'peat drainage', 'tropospheric',
        #     'permafrost', 'centennial', 'aerosols', 'sequestration', 'albedo', 'exacerbating']
        assert words == ['anthropogenic',
 'tropospheric',
 'permafrost',
 'sequestration',
 'albedo',
 'anthropogenic',
 'physical-biogeochemical',
 'peat drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']

        # list of files
        words = FileLib.get_input_strings([
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_5.txt"),
            Path(Resources.TEST_RESOURCES_DIR, "wordlists", "small_10.txt")], split=True)

        assert words == ['anthropogenic',
 'tropospheric',
 'permafrost',
 'sequestration',
 'albedo',
 'anthropogenic',
 'physical-biogeochemical',
 'peat',
 'drainage',
 'tropospheric',
 'permafrost',
 'centennial',
 'aerosols',
 'sequestration',
 'albedo',
 'exacerbating']




def main():
    File0Test.test_file_simple()
    File0Test.test_templates()



"""classmethods from headless tests. May be useful elsewhere

"""
import csv
import re
from collections import Counter
from pathlib import Path
import lxml.etree as ET
from lxml.html import HTMLParser

from amilib.ami_html import HtmlLib
from amilib.file_lib import FileLib
from amilib.util import Util
from amilib.wikimedia import WikidataLookup
from amilib.xml_lib import HtmlElement, XmlLib

logger = Util.get_logger(__name__)


class HeadlessLib:

    @classmethod
    def predict_encoding(cls, file_path: Path, n_lines: int = 20) -> str:
        import chardet
        '''Predict a file's encoding using chardet'''

        # Open the file as binary data
        with Path(file_path).open('rb') as f:
            # Join binary lines for specified number of lines
            rawdata = b''.join([f.readline() for _ in range(n_lines)])

        return chardet.detect(rawdata)['encoding']

    @classmethod
    def edit_title(cls, dict_html):
        """edit main title, extracting brackets and guillemets"""
        h4 = dict_html.xpath(".//h4[contains(@class, 'bg-primary')]")
        if len(h4) == 1:
            title = h4[0].text
            parent = h4[0].getparent()
            title = cls.extract_chunks(title, "(.*)«([^»]*)»(.*)", parent, "wg")
            title = cls.extract_chunks(title, "(.*)\\(([^\\)]*)\\)(.*)", parent, "paren")
            title = title.strip()
            # lowercase unless string has embedded uppercase
            if sum(i.isupper() for i in title[1:]) == 0:
                title = title.lower()
            h4[0].text = title
            return title

    @classmethod
    def extract_chunks(cls, text, regex, parent, tag, count=999):
        """extracts inline chunks and closes up the text
        can be be iterative
        creates <div> children of <parent> and labels them with a class/tag attribute
        then adds chunks
        :param text: chunk to analyse
        :param regex: has 3 groups (pre, chunk, post)
        :param paraent: to add results to
        :param count: number of times to repeat (default = 999)
        :return: de-chunked text"""
        t = text
        while True:
            match = re.match(regex, t)
            if not match:
                return t
            wg = ET.SubElement(parent, "div")
            wg.attrib["class"] = tag
            wg.text = match.group(2)
            t = match.group(1) + match.group(3)
        return t

    @classmethod
    def analyze_parent(cls, h6):
        p_string = """
    <div class="p-3 small">
    <h6 class="fs-6">Parent-term</h6>
    <ul class="items text-muted fs-6"><li><span class="specificlink" data-report="AR6" data-phrase="Mass balance/budget (of glaciers or ice sheets)" data-phraseid="2202">Mass balance/budget (of glaciers or ice sheets)</span></li></ul>
    </div>
    """
        id_phrases = cls.extract_id_phrases(h6)
        if len(id_phrases) != 1:
            raise Exception(f"expected 1 parent, found {id_phrases}")
        return id_phrases[0]

    @classmethod
    def extract_id_phrases(cls, h6):
        """"""

        lis = h6.xpath("../ul/li")
        id_phrases = []
        for li in lis:
            span = li.xpath("span")[0]
            phrase = span.text
            data_phrase = span.attrib["data-phrase"]
            if phrase != data_phrase:
                logger.info(f"phrase {phrase} != data_phrase {data_phrase}")
            phrase_id = span.attrib["data-phraseid"]
            id_phrases.append((phrase_id, phrase))
        return id_phrases

    @classmethod
    def analyze_sub_terms(cls, entry_html):
        """
        <div class="ms-2 p-1 small text-muted">
    <h6 class="mb-0">Sub-terms</h6>
    <ul class="items mb-0">
    <li><span class="specificlink" data-report="AR6" data-phrase="Agricultural and ecological drought" data-phraseid="5559">Agricultural and ecological drought</span></li>
    <li><span class="specificlink" data-report="AR6" data-phrase="Hydrological drought" data-phraseid="5557">Hydrological drought</span></li>
    <li><span class="specificlink" data-report="AR6" data-phrase="Megadrought" data-phraseid="209">Megadrought</span></li>
    <li><span class="specificlink" data-report="AR6" data-phrase="Meteorological drought" data-phraseid="5587">Meteorological drought</span></li>
    </ul>
    </div>
        :return: list of (id, term) triples
        """
        id_phrases = cls.extract_id_phrases(entry_html)
        if len(id_phrases) == 0:
            raise Exception(f"expected 1 or more sub-terms, found {id_phrases}")
        return id_phrases

    @classmethod
    def analyze_references(cls, entry_html):
        """
        the text is messy. Seems to be
        - text <br>
        - text <br>
        <div class="ms-2 p-1 small text-muted">
    <h6 class="mb-0">References</h6> - SpanMarker, 2021: Reporting and accounting of LULUCF activities under the Kyoto Protocol. United Nations Framework Convention on Climate Change (SpanMarker), Bonn, Germany. Retrieved from: https://unfccc.int/topics/land-use/workstreams/land-use-land-use-change-and-forestry-lulucf/reporting-and-accounting-of-lulucf-activities-under-the-kyoto-protocol<br> - SpanMarker, 2021: Reporting and Review under the Paris Agreement. United Nations Framework Convention on Climate Change (SpanMarker), Bonn, Germany. Retrieved from: https://unfccc.int/process-and-meetings/transparency-and-reporting/reporting-and-review-under-the-paris-agreement<br>
    </div>
        """
        divs = entry_html.xpath(".//div[h6]")
        ld = len(divs) > 1
        for div in divs:
            if ld:
                # logger.info(f" DIV: {lxml.etree.tostring(div, pretty_print=True)}")
                pass
            nodes = div.xpath("./node()")
            texts = []
            text = "NONE"
            for node in nodes:
                if type(node) is HtmlElement:
                    if node.tag == "h6":
                        pass
                    elif node.tag == "br":
                        texts.append(text)
                    else:
                        raise Exception(f"unexpected tag: {node.tag}")
                elif type(node) is ET._ElementUnicodeResult:
                    text = str(node)
                else:
                    # logger.info(f":text {node}")
                    raise Exception(f" unknown node {type(node)}")
                if len(texts) == 0:
                    # logger.info("NO TEXTS")
                    pass
                else:
                    logger.info(f"texts: {len(texts)}:: {texts}")

    @classmethod
    def remove_styles(cls, entry_html):
        """

        """
        style_elems = entry_html.xpath(".//*[@style]")
        for style_elem in style_elems:
            style_elem.attrib["style"] = None

    @classmethod
    def edit_paras(cls, entry_html):
        """

        """
        # TODO fix regex to find missinf first sentences
        regex = re.compile("(.)\\s+(.*)")
        # this may not be universal
        mainclass = "fs-6 p-2 mb-0"
        ps = entry_html.xpath(f"//p")
        for p in ps:
            if p.text is None:
                continue
            # if p.text.startswith("A change in functional or"):
            #     logger.info(f"CHANGE")
            clazz = p.attrib.get('class')
            if clazz == mainclass:
                # this is crude; split first sentence into 2 paras
                text = ET.tostring(p, encoding=str)
                # find period to split sentence
                # TODO some paragraphs are not split
                # match = re.match(regex, s)
                split = re.split("\\.\\s+", text, 1)
                if len(split) == 1:
                    cls.make_definition_para(p)
                else:
                    div = None
                    for tag in ["p", "span"]:
                        ss = "<div>" + split[0] + "." + f"</{tag}><{tag}>" + split[1] + f"</div>"
                        # reparse; there may be subelements such as span
                        try:
                            div = ET.fromstring(ss)
                            break
                        except Exception as e:
                            pass
                    if div is None:
                        logger.warning(f"FAIL {ss}")
                        continue
                    p.getparent().replace(p, div)
                    p0 = div.xpath("./p")[0]
                    cls.make_definition_para(p0)

    @classmethod
    def make_definition_para(cls, p):
        p.attrib["style"] = "font-weight: bold"
        p.attrib["class"] = "definition"

    @classmethod
    def edit_lists(cls, entry_html, parent_id_set=None, subterm_id_set=None):
        """div class="p-3 small"><h6 class="fs-6">Parent-term<"""

        # h6s = entry_html.xpath(f".//div[@class='p-3 small']/h6")
        dh6s = entry_html.xpath(f".//div[h6]")
        if len(dh6s) == 0:
            # logger.warning(f"No div/h6 found")
            return
        for dh6 in dh6s:
            h6 = dh6.xpath("./h6")[0]
            txt = h6.text.strip()
            if txt == "Parent-term":
                parent_id = cls.analyze_parent(h6)
                parent_id_set.add(parent_id)
            elif txt == "Sub-terms":
                subterm_id_list = cls.analyze_sub_terms(h6)
                for id_phrase in subterm_id_list:
                    subterm_id_set.add(id_phrase[0])
            elif txt == "References":
                cls.analyze_references(h6)
            else:
                raise Exception(f"unknown list title {txt}")

        """
    
        <div class="ms-2 p-1 small text-muted">
    <h6 class="mb-0">Sub-terms</h6>
    <ul class="items mb-0"><li><span class="specificlink" data-report="AR6" data-phrase="Household carbon footprint" data-phraseid="5376">Household carbon footprint</span></li></ul>
    </div>
    <div class="ms-2 p-1 small text-muted">
    <h6 class="mb-0">References</h6> - Wiedmann, T. and Minx, J. C. (2008). A definition of carbon footprint, in C. Pertsova (ed.), Ecological Economics Research Trends, Nova Science Publishers, Hauppauge NY, chapter 1, pp. 1â11. URL: https://www.novapublishers.com/catalog/product info.php?products id=5999<br>
    </div>"""

    @classmethod
    def make_targets(cls, text):
        """creates a list to search with
        if text ends with 's' returns ['foos', 'foo']"""
        ss = []
        if text is not None:
            text = cls.make_title_id(text)
            ss.append(text)
            if text[-1:] == 's':
                ss.append(text[:-1])
        return ss

    @classmethod
    def markup_em_and_write_files(cls, entry_html_list, entry_by_id):
        """create a Counter of <em> to see which might be terms"""
        em_counter = Counter()
        missing_targets = set()
        for entry in entry_html_list:
            entry_html = entry[0]
            out_path = entry[1]
            name = entry_html.xpath(".//body/a/@name")
            name = name[0] if len(name) > 0 else ""
            # TODO include parent/subterms
            cls.find_mentions(em_counter, entry_by_id, entry_html, missing_targets)
            HtmlLib.write_html_file(entry_html, out_path, debug=True)
        return missing_targets, em_counter

    @classmethod
    def find_mentions(cls, em_counter, entry_by_id, entry_html, missing_targets):
        """TODO Badly need a class to manage this"""
        ems = entry_html.xpath(".//em")
        for em in ems:
            text = cls.make_title_id(em.text)
            em_targets = cls.make_targets(text)
            matched = None
            for em_target in em_targets:
                matched0 = cls.match_target_in_dict(em_target, entry_by_id)
                matched = matched0 if matched0 is not None else matched
            if matched:
                em_counter[em.text] += 1
                a = ET.SubElement(em, "a")
                a.attrib["href"] = "#" + em_target
                a.attrib["class"] = "mention"
                a.text = em.text
                em.text = ""
            else:
                missing_targets.add(em.text)

    @classmethod
    def find_wikidata(cls, entry_html):
        term = entry_html.xpath("//a/name")[0]
        term = term.replace("_", " ").strip()
        logger.info(f"term {term}")

        wikidata_lookup = WikidataLookup()
        qitem0, desc, wikidata_hits = wikidata_lookup.lookup_wikidata(term)
        logger.info(f"qitem {qitem0}")

    @classmethod
    def match_target_in_dict(cls, em_target, entry_by_id):
        target_id = cls.make_title_id(em_target)
        match = entry_by_id.get(target_id)
        if match is not None:
            # logger.info(f"MATCHED {match}")
            pass
        return match is not None

    @classmethod
    def extract_term_from_title(cls, entry_html):
        """
        <h4 class="fw-bold fs-5 bg-primary text-light p-2">Aerosol effective radiative forcing (ERFari+aci)  « WGI »</h4>
        NYI
        """
        h4_fs_5 = entry_html.xpath(".//h4[contains(@class,'fs-5') and contains(@class, 'fw-bold')]")

    @classmethod
    def make_title_id(cls, title):
        if title is None:
            return None
        # strip brackets
        match = re.match("(.*)\\(.*", title)
        if match:
            title = match.group(1)
        title_id = title.strip().replace(" ", "_").lower()
        return title_id

    @classmethod
    def write_missing(cls, missing_targets, filename):
        targets = [t for t in missing_targets if t is not None]
        with open(filename, "w") as f:
            for t in sorted(targets):
                f.write(t + "\n")

    @classmethod
    def extract_mention_links(cls, entry_html_list, filename):
        with open(filename, 'w') as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(["source", "role", "target", ])
            for (entry_html, _) in entry_html_list:
                name = entry_html.xpath(".//a/@name")[0]
                refs = entry_html.xpath(".//a[@class='mention']")
                for ref in refs:
                    href = ref.attrib["href"][1:]  # first char is hash
                    csvwriter.writerow([name, "mentions", href, ])

    @classmethod
    def extract_parent_subterms(cls, entry_html_list, filename):
        with open(filename, 'w') as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(["source", "role", "target", ])
            for role_name, role in [("Parent-term", "parent"), ("Sub-terms", "subterm")]:
                for (entry_html, _) in entry_html_list:
                    name = entry_html.xpath(".//a/@name")[0]
                    refs = entry_html.xpath(f".//div[h6[.='{role_name}']]/ul/li/span")
                    # TODO move to earlier
                    for ref in refs:
                        if ref.text is None:
                            logger.info(f"Null text for name {name}")
                            continue
                        ref_id = cls.make_title_id(ref.text)
                        a = ET.SubElement(ref, "a")
                        a.attrib["href"] = "#" + ref_id
                        a.text = ref.text
                        ref.text = ""
                        csvwriter.writerow([name, role, ref_id, ])

    @classmethod
    def make_glossary(cls, dict_files, out_dir, total_gloss_dir=None, total_glossary=None, debug=True):
        titles = set()
        parent_id_set = set()
        subterm_id_set = set()
        encoding = "UTF-8"
        entry_by_id = dict()
        entry_html_list = []
        for dict_file in dict_files:
            entry_html = ET.parse(str(dict_file), parser=HTMLParser(encoding=encoding)).getroot()
            # remove "Cloae" button
            XmlLib.remove_all(entry_html, ".//div[@class='modal-footer']", debug=False)
            # remove "AR6" button
            XmlLib.remove_all(entry_html, ".//h5[button]", debug=False)

            title = cls.edit_title(entry_html)
            title_id = cls.make_title_id(title)
            dict_body = HtmlLib.get_body(entry_html)
            # html anchor for every element
            entry_a = ET.SubElement(dict_body, "a")
            entry_a.attrib["name"] = title_id
            # are there duplicate titles after trimming and lowercasing
            if entry_by_id.get(title_id) is not None:
                logger.info(f"duplicate title_id {title_id}")
                continue
            entry_by_id[title_id] = entry_html

            cls.remove_styles(entry_html)
            cls.extract_term_from_title(entry_html)
            cls.edit_paras(entry_html)
            cls.edit_lists(entry_html, parent_id_set=parent_id_set, subterm_id_set=subterm_id_set)
            titles.add(title)

            # output
            html_out = HtmlLib.create_html_with_empty_head_body()
            HtmlLib.add_charset(html_out)
            body = HtmlLib.get_body(html_out)

            body.getparent().replace(body, dict_body)
            a = ET.SubElement(dict_body, "a")
            a.attrib["name"] = title_id

            path = cls.create_out_path(dict_file, out_dir)
            if not path:
                continue
            entry_html_list.append((html_out, path))
        logger.info(f"parent: {len(parent_id_set)} {parent_id_set}")
        logger.info(f"parent: {len(subterm_id_set)} {subterm_id_set}")

        logger.info(f"Must fix to write the modified HTML file")

        cls.extract_mention_links(entry_html_list, Path(total_glossary, "mentions.csv"))
        cls.extract_parent_subterms(entry_html_list, Path(total_glossary, "parents.csv"))
        missing_targets, em_counter = cls.markup_em_and_write_files(entry_html_list, entry_by_id)
        cls.write_missing(missing_targets, Path(total_glossary, "missing_em_targets.txt"))
        logger.info(f"entry_dict {len(entry_by_id)}")
        gloss_ids_file = str(Path(total_glossary, "ids.txt"))
        with open(gloss_ids_file, "w") as f:
            for key in sorted(entry_by_id.keys()):
                entry = entry_by_id.get(key)
                f.write(f"{key} {entry.xpath('/html/body/a/@name')}\n")
        if debug:
            logger.info(f"wrote {gloss_ids_file}")
        logger.warning(f"missing targets: {len(missing_targets)} {missing_targets}")
        logger.info(f"em_counter {len(em_counter)} {em_counter}")

    @classmethod
    def make_header(cls, tr):
        th = ET.SubElement(tr, "th")
        th.text = "input"
        tr.append(th)
        th = ET.SubElement(tr, "th")
        th.text = "output"
        tr.append(th)

    @classmethod
    def make_cell(cls, file, output_name, tr, style=None, filename=False):
        td = ET.SubElement(tr, "td")
        td.attrib["style"] = "padding : 4px; margin : 4px; background : #fee;"
        if (filename):
            h3 = ET.SubElement(td, "h3")
            h3.text = output_name
        # html in output glossary
        try:
            body = ET.parse(str(file), parser=HTMLParser()).xpath("//body")[0]
        except Exception as e:
            logger.warning(f"failed to parse {file} giving {e}")
            return
        a = body.xpath("./a")
        divtop = ET.parse(str(file), parser=HTMLParser()).xpath("//body/div")[0]
        if len(a) > 0:
            divtop.insert(0, a[0])
        div = ET.SubElement(td, "div")
        if True or not style:
            style = "margin : 8px; padding : 8px; background : #eee;"
        div.attrib["style"] = style

        div.append(divtop)

    @classmethod
    def create_out_path(cls, dict_file, out_dir):
        path = dict_file
        stem0 = Path(dict_file).stem
        match = re.match("(.+)_(?:[A-Z]|123)$", stem0)
        if match:
            stem = match.group(1)
            path = Path(out_dir, f"{stem}.html")
        return path



from pathlib import Path

from amilib.amix import AmiLib
from amilib.util import Util
from test.resources import Resources
from test.test_all import AmiAnyTest

logger = Util.get_logger(__name__)

class SearchTest(AmiAnyTest):
    pass

    def test_help(self):
        logger.debug("do nothing")
        args = ([
            "SEARCH",
            "--help",
            ])
        pyami = AmiLib()
        pyami.run_command(args)

    def test_annotate_file_with_phrases(self):
        """
        read inpath HTML, and use dict to match words and phrases
        """
        stem = "carbon_cycle1"
        inpath = Path(Resources.TEST_RESOURCES_DIR, "ipcc", "wg3", "Chapter03", "html_with_ids.html")
        dictpath = Path(Resources.TEST_RESOURCES_DIR, "dictionary", "climate", f"{stem}.html")
        outpath = Path(Resources.TEMP_DIR, "annotate", f"{stem}.html")
        args = ([
            "SEARCH",
            "--inpath", inpath,
            "--dict", dictpath,
            "--outpath", outpath,
            "--operation", "annotate"

            ])
        pyami = AmiLib()
        pyami.run_command(args)
        assert outpath.exists(), f"{outpath} should exist"

    def test_remove_styles_annotate_file_with_phrases(self):
        """
        read inpath HTML, remove styles, and use dict to match words and phrases
        """
        stem = "carbon_cycle1"
        inpath = Path(Resources.TEST_RESOURCES_DIR, "ipcc", "wg3", "Chapter03", "html_with_ids.html")
        dictpath = Path(Resources.TEST_RESOURCES_DIR, "dictionary", "climate", f"{stem}.html")
        outpath = Path(Resources.TEMP_DIR, "annotate", f"{stem}.html")
        args = ([
            "SEARCH",
            "--inpath", inpath,
            "--dict", dictpath,
            "--outpath", outpath,
            "--operation", "annotate", "no_input_styles",

            ])
        pyami = AmiLib()
        pyami.run_command(args)
        assert outpath.exists(), f"{outpath} should exist"





import ast
import collections
import configparser
import json
import logging
from pathlib import Path
import pandas as pd

from amilib.ami_bib import (SAVED, SAVED_CONFIG_INI, SECTION_KEYS, API, LIMIT, QUERY, STARTDATE, XML, \
                            EUPMC_RESULTS_JSON, PMCID, ABS_TEXT, EPMC_KEYS, JOURNAL_INFO, DOI, TITLE, AUTHOR_STRING,
                            PUB_YEAR, JOURNAL_INFO_TITLE, Pygetpapers)
from amilib.ami_html import HtmlUtil, HtmlLib
from amilib.ami_util import AmiJson, AmiUtil
from amilib.ami_corpus import AmiCorpus
from amilib.amix import AmiLib
from amilib.util import Util
from test.resources import Resources
from test.test_all import AmiAnyTest

logger = Util.get_logger(__name__)

class AmiBibliographyTest(AmiAnyTest):
    """

    """
    @classmethod
    def convert_csv_to_bib(cls):
        """

        """

def df_toupper(s):
    return f"<a href='{s}`>s</a>"

def df_truncate(s):
    return f"{s[:200]}..."

class DictParser(ast.NodeVisitor):
    def visit_Dict(self,node):
        keys,values = node.keys,node.values
        keys = [n.s for n in node.keys]
        values = [n.s for n in node.values]
        self.od = collections.OrderedDict(zip(keys, values))

def df_unpack_dict(json_string):
    """
    this is messy. I think we need a recursive descent parser
    """
    print(f"DIKT {json_string}")
    try:
        # dikt = json.loads(s)
    # or
        # decoder = json.JSONDecoder(object_pairs_hook=collections.OrderedDict)
        # dikt = decoder.decode(json_string)
    # or
        dp = DictParser()
        dp.visit(ast.parse(json_string))
        dikt = dp.od
    except Exception as e:
        print(f"EXC {e}")
        return "None"
    print(f"dikt {type(dikt)} {dikt.keys()}")
    title = dikt.get("journal").get("title")
    return title

class PygetpapersTest(AmiAnyTest):
    """
    manage output from pygetpapers
    """

    def test_edit_csv(self):
        """
        reade *.csv created by pygetpapers
        not working because of lack of OrderedDict parser
        """
        infile = Path(Resources.TEST_RESOURCES_DIR, "csv", "frictionless", "europe_pmc.csv")
        outdir = Path(Resources.TEMP_DIR, "csv", "frictionless")
        outdir.mkdir(parents=True, exist_ok=True)
        outfile = Path(outdir, "europe_pmc.csv")
        outfile_h = Path(outdir, "europe_pmc.html")
        assert infile.exists()
        df = pd.read_csv(infile)

        keys = df.keys()
        key_list = keys.to_list()
        assert EPMC_KEYS == keys.to_list()
        df[PMCID] = df[PMCID].apply(df_toupper)
        df[ABS_TEXT] = df[ABS_TEXT].apply(df_truncate)
        df[JOURNAL_INFO] = df[JOURNAL_INFO].apply(df_unpack_dict)
        df2 = df[[PMCID, DOI, TITLE, AUTHOR_STRING, JOURNAL_INFO, PUB_YEAR, ABS_TEXT]]

        with open(outfile, "w") as f:
            f.write(df2.to_csv())
        with open(outfile_h, "w") as f:
            f.write(df2.to_html(escape="False"))

    def test_make_html_table_from_json(self):
        """
        assume JSON is an implicit table
        """
        effective_level = logger.getEffectiveLevel()
        logger.setLevel(logging.INFO)

        project_name = "district_heating"
        indir = Path(Resources.TEST_RESOURCES_DIR, "json", project_name)
        outdir = Path(Resources.TEMP_DIR, "json", project_name)
        outfile_h = Path(outdir, "europe_pmc.html")

        AmiCorpus.make_datatables(indir, outdir, outfile_h)
        logger.setLevel(effective_level)

    def test_make_datatables_cli(self):
        """

        """
        indir = Path(Resources.TEST_RESOURCES_DIR, "json", "district_heating")
        args = ["HTML", "--indir", str(indir), "--operation", "DATATABLES"]
        amilib = AmiLib()
        amilib.run_command(args)

    def test_read_pygetpapers_config(self):
        """
        reads pygetpapers saved_config.ini and extracts meta/data
        """
        inpath = Path(Resources.TEST_RESOURCES_DIR, "pygetpapers", SAVED_CONFIG_INI)
        assert inpath.exists()
        saved_section, section_names = Pygetpapers.get_saved_section(inpath)

        assert type(section_names) is list
        assert section_names == [SAVED]
        assert type(saved_section) is configparser.SectionProxy

        assert SECTION_KEYS == saved_section.keys()
        assert saved_section.get(API) == "europe_pmc"
        assert saved_section.get(LIMIT) == '20'
        assert saved_section.get(QUERY) == "'district heating'"
        assert saved_section.get(STARTDATE) == "False"
        assert saved_section.get(XML) == "True"


